# coding: utf-8

from platform import system
from multiprocessing import Process, Value, Barrier, Event, Queue, \
  get_start_method, synchronize, queues
from multiprocessing.sharedctypes import Synchronized
from multiprocessing.connection import wait
from threading import BrokenBarrierError, Thread
from queue import Empty
import logging
import logging.handlers
from time import sleep, time, time_ns
from weakref import WeakSet
from typing import Union, Optional, List, Dict, Any, Iterable
from collections import defaultdict
import subprocess
from sys import stdout, stderr, argv
from pathlib import Path

from .meta_block import MetaBlock
from ...links import Link
from ..._global import LinkDataError, StartTimeout, PrepareError, \
  T0NotSetError, GeneratorStop, ReaderStop, CameraPrepareError, \
  CameraRuntimeError, CameraConfigError, CrappyFail
from ...tool.ft232h import USBServer


class Block(Process, metaclass=MetaBlock):
  """This class constitutes the base object in Crappy.

  It is extremely versatile, and can perform a wide variety of actions during a
  test. Many Blocks are already defined in Crappy, but it is possible to define
  custom ones for specific purposes.

  It is a subclass of :obj:`multiprocessing.Process`, and is thus an
  independent process in Python. It communicates with other Blocks via
  :mod:`multiprocessing` objects.

  This class also contains the class methods that allow driving a script with
  Crappy. They are always called in the `__main__` Process, and drive the
  execution of all the children Blocks.
  
  .. versionadded:: 1.4.0
  """

  instances = WeakSet()
  names: List[str] = list()
  log_level: Optional[int] = logging.DEBUG

  # The synchronization objects will be set later
  shared_t0: Optional[Synchronized] = None
  ready_barrier: Optional[synchronize.Barrier] = None
  start_event: Optional[synchronize.Event] = None
  stop_event: Optional[synchronize.Event] = None
  raise_event: Optional[synchronize.Event] = None
  kbi_event: Optional[synchronize.Event] = None
  logger: Optional[logging.Logger] = None
  log_queue: Optional[queues.Queue] = None
  log_thread: Optional[Thread] = None
  thread_stop: bool = False
  no_raise: bool = False

  prepared_all: bool = False
  launched_all: bool = False

  def __init__(self) -> None:
    """Sets the attributes and initializes the parent class."""

    super().__init__()

    # The lists of input and output links
    self.outputs: List[Link] = list()
    self.inputs: List[Link] = list()

    # Various objects that should be set by child classes
    self.niceness: int = 0
    self.labels: Optional[Iterable[str]] = None
    self.freq = None
    self.display_freq = False
    self.name = self.get_name(type(self).__name__)

    # The synchronization objects will be set later
    self._instance_t0: Optional[Synchronized] = None
    self._ready_barrier: Optional[synchronize.Barrier] = None
    self._start_event: Optional[synchronize.Event] = None
    self._stop_event: Optional[synchronize.Event] = None
    self._raise_event: Optional[synchronize.Event] = None
    self._kbi_event: Optional[synchronize.Event] = None

    # The objects for logging will be set later
    self._log_queue: Optional[queues.Queue] = None
    self._logger: Optional[logging.Logger] = None
    self._debug: Optional[bool] = False
    self._log_level: int = logging.INFO

    # Objects for displaying performance information about the block
    self._last_t: Optional[float] = None
    self._last_fps: Optional[float] = None
    self._n_loops: int = 0

    self._last_values = None

  def __new__(cls, *args, **kwargs):
    """Called when instantiating a new instance of a Block.

    Adds itself to the :obj:`~weakref.WeakSet` listing all the instantiated
    Blocks.
    """

    instance = super().__new__(cls)
    cls.instances.add(instance)
    return instance

  @classmethod
  def get_name(cls, name: str) -> str:
    """Method attributing to each new Block a unique name, based on the name of
    the class and the number of existing instances for this class.
    
    .. versionadded:: 2.0.0
    """

    i = 1
    while f"crappy.{name}-{i}" in cls.names:
      i += 1

    cls.names.append(f"crappy.{name}-{i}")
    return f"crappy.{name}-{i}"

  @classmethod
  def start_all(cls,
                allow_root: bool = False,
                log_level: Optional[int] = logging.DEBUG,
                no_raise: bool = False) -> None:
    """Method for starting a script with Crappy.

    It sets the synchronization objects for all the Blocks, renices the
    corresponding :obj:`~multiprocessing.Process` and starts the Blocks.

    The call to this method is blocking until Crappy finishes.

    Note:
      It is possible to have a finer grained control of the start of a Crappy
      script with the methods :meth:`prepare_all`, :meth:`renice_all` and
      :meth:`launch_all`.

    Args:
      allow_root: If set to :obj:`True`, tries to renice the Processes with
        sudo privilege in Linux. It requires the Python script to be run
        with sudo privilege, otherwise it has no effect.
      log_level: The maximum logging level that will be handled by Crappy. By
        default, it is set to the lowest level (:obj:`~logging.DEBUG`) so that
        all messages are handled. If set to a higher level, the levels
        specified for each Block with the ``debug`` argument may be ignored. If
        set to :obj:`None`, logging is totally disabled. Refer to the
        documentation of the :mod:`logging` module for information on the
        possible levels.
      no_raise: When set to :obj:`False`, the Exceptions encountered during
        Crappy's execution, as well as the :exc:`KeyboardInterrupt`, will raise
        an Exception right before Crappy returns. This is meant to prevent the
        execution of code that would come after Crappy, in case Crappy does not
        terminate as expected. This behavior can be disabled by setting this
        argument to :obj:`True`.
    
    .. versionchanged:: 2.0.0 renamed *high_prio* argument to *allow_root*
    .. versionremoved:: 2.0.0 *t0*, *verbose*, *bg* arguments
    .. versionadded:: 2.0.0 *log_level*, *no_raise* arguments
    """

    cls.prepare_all(log_level)
    cls.renice_all(allow_root)
    cls.launch_all(no_raise)

  @classmethod
  def prepare_all(cls, log_level: Optional[int] = logging.DEBUG) -> None:
    """Creates the synchronization objects, shares them with the Blocks, and
    starts the :obj:`~multiprocessing.Process` associated to the Blocks.

    Also initializes the :obj:`~logging.Logger` for the Crappy script.

    Once started with this method, the Blocks will call their :meth:`prepare`
    method and then be blocked by a :obj:`multiprocessing.Barrier`.

    If an error is caught at a moment when the Blocks might already be running,
    performs an extensive cleanup to ensure everything stops as expected.

    Args:
      log_level: The maximum logging level that will be handled by Crappy. By
        default, it is set to the lowest level (:obj:`~logging.DEBUG`) so that 
        all messages are handled. If set to a higher level, the levels 
        specified for each Block with the ``debug`` argument may be ignored. If 
        set to :obj:`None`, logging is totally disabled. Refer to the 
        documentation of the :mod:`logging` module for information on the 
        possible levels.
    
    .. versionremoved:: 2.0.0 *verbose* argument
    .. versionadded:: 2.0.0 *log_level* argument
    """

    # Flag indicating whether to perform the cleanup action or not
    cleanup = False

    try:
      # Making sure that the Block classmethods are called in the right order
      if cls.prepared_all:
        cls.cls_log(logging.ERROR,
                    "The method prepare_all was already called ! This is "
                    "unexpected, aborting !")
        # As Crappy was already initialized, it must now be cleaned up
        cleanup = True
        # Raising will skip all the setup part and keep the existing context
        raise RuntimeError

      if cls.launched_all:
        cls.cls_log(logging.ERROR, "The launched_all flag is unexpectedly "
                                   "raised, aborting !")
        # As Crappy was already initialized, it must now be cleaned up
        cleanup = True
        # Raising will skip all the setup part and keep the existing context
        raise RuntimeError

      cls.log_level = log_level

      # Initializing the logger and displaying the first messages
      cls._set_logger()
      cls.cls_log(logging.INFO,
                  "===================== CRAPPY =====================")
      cls.cls_log(logging.INFO, f'Starting the script {argv[0]}\n')
      cls.cls_log(logging.INFO, 'Logger configured')

      # Setting all the synchronization objects at the class level
      cls.ready_barrier = Barrier(len(cls.instances) + 1)
      cls.shared_t0 = Value('d', -1.0)
      cls.start_event = Event()
      cls.stop_event = Event()
      cls.raise_event = Event()
      cls.kbi_event = Event()
      cls.cls_log(logging.INFO, 'Multiprocessing synchronization objects set '
                                'for main process')

      # Starting from that point, Crappy has to be cleaned up if anything wrong
      # happens
      cleanup = True

      # Initializing the objects required for logging
      cls.log_queue = Queue()
      cls.log_thread = Thread(target=cls._log_target)
      if get_start_method() == 'spawn':
        cls.log_thread.start()
        cls.cls_log(logging.INFO, 'Logger thread started')

      # Starting the USB server if required
      if USBServer.initialized:
        cls.cls_log(logging.INFO, "Starting the USB server")
        USBServer.start_server(cls.log_queue, logging.INFO)

      # Passing the synchronization and logging objects to each Block
      for instance in cls.instances:
        instance._ready_barrier = cls.ready_barrier
        instance._instance_t0 = cls.shared_t0
        instance._stop_event = cls.stop_event
        instance._start_event = cls.start_event
        instance._raise_event = cls.raise_event
        instance._kbi_event = cls.kbi_event
        instance._log_queue = cls.log_queue
        cls.cls_log(logging.INFO, f'Multiprocessing synchronization objects '
                                  f'set for {instance.name} Block')

        # Setting the common log level to all the instances
        if instance._log_level is not None:
          if cls.log_level is not None:
            instance._log_level = max(instance._log_level, cls.log_level)
          else:
            instance._log_level = None
        cls.cls_log(logging.INFO, f"Log level set for the {instance.name} "
                                  f"Block")

      # Starting all the Blocks
      for instance in cls.instances:
        instance.start()
        cls.cls_log(logging.INFO, f'Started the {instance.name} Block')

      # Setting the prepared flag
      cls.prepared_all = True

    # At that point the Blocks might be started or not. If started, they are
    # preparing or waiting at the Barrier
    except (Exception, KeyboardInterrupt) as exc:

      # If there is no specific cleanup to perform, only raising
      if not cleanup:
        raise

      # KeyboardInterrupt is a separate case
      if isinstance(exc, KeyboardInterrupt):
        cls.cls_log(logging.WARNING, "Caught KeyboardInterrupt in the main "
                                     "Process while running prepare_all !")
        # Special Event for the KeyboardInterrupt
        cls.kbi_event.set()
        cls.cls_log(logging.WARNING, 'Set the KbI Event after catching '
                                     'KeyboardInterrupt in the main Process '
                                     'in prepare_all')
      # General case
      else:
        cls.logger.exception("Caught exception while running prepare_all, "
                             "aborting", exc_info=exc)
        # Any Exception caught in the main Process must stop the script
        cls.raise_event.set()
        cls.cls_log(logging.WARNING, 'Set the raise Event after exception was '
                                     'caught in the main Process in '
                                     'prepare_all')
      # Breaking the Barrier to warn other Processes that something went wrong
      cls.ready_barrier.abort()
      cls.cls_log(logging.WARNING, "Broke the Barrier due to an exception "
                                   "caught in prepare_all")
      # Need to clean up as some Blocks might already be running
      cls._cleanup()

  @classmethod
  def renice_all(cls, allow_root: bool) -> None:
    """On Linux and macOS, renices the :obj:`~multiprocessing.Process` 
    associated with the Blocks.

    On Windows, does nothing.

    If an error is caught, performs an extensive cleanup to ensure everything
    stops as expected.

    Args:
      allow_root: If set to :obj:`True`, tries to renice the Processes with 
        sudo privilege in Linux. It requires the Python script to be run with 
        sudo privilege, otherwise it has no effect.
    
    .. versionchanged:: 2.0.0 renamed *high_prio* argument to *allow_root*
    """

    # Flag indicating whether to perform the cleanup action or not
    cleanup = True

    try:
      # Making sure that the Block classmethods are called in the right order
      if not cls.prepared_all:
        cls.cls_log(logging.ERROR, "Cannot call renice before calling "
                                   "prepare ! Aborting")
        # If prepare wasn't called, there is no need to clean up Crappy
        cleanup = False
        raise RuntimeError("Cannot call renice before calling prepare ! "
                           "Aborting")

      if cls.launched_all:
        cls.cls_log(logging.ERROR, "The launched_all flag is unexpectedly "
                                   "raised, aborting !")
        raise RuntimeError

      # There's no niceness on Windows
      if system() == "Windows":
        cls.cls_log(logging.INFO, 'Not renicing processes on Windows')
        return

      # Renicing all the Blocks
      cls.cls_log(logging.INFO, 'Renicing processes')
      for inst in cls.instances:
        # If root is not allowed then the minimum niceness is 0
        niceness = max(inst.niceness, 0 if not allow_root else -20)

        # System call for setting the niceness
        if niceness < 0:
          subprocess.call(['sudo', 'renice', str(niceness), '-p',
                           str(inst.pid)], stdout=subprocess.DEVNULL)
          cls.cls_log(logging.INFO, f"Reniced process {inst.name} with PID "
                                    f"{inst.pid} to niceness {niceness} "
                                    f"with sudo privilege")
        else:
          subprocess.call(['renice', str(niceness), '-p', str(inst.pid)],
                          stdout=subprocess.DEVNULL)
          cls.cls_log(logging.INFO, f"Reniced process {inst.name} with PID "
                                    f"{inst.pid} to niceness {niceness}")

    # At that point the Blocks should be preparing or waiting at the Barrier
    except (Exception, KeyboardInterrupt) as exc:

      # If there is no specific cleanup to perform, only raising
      if not cleanup:
        raise

      # KeyboardInterrupt is a separate case
      if isinstance(exc, KeyboardInterrupt):
        cls.cls_log(logging.WARNING, "Caught KeyboardInterrupt in the main "
                                     "Process while running renice_all !")
        # Special Event for the KeyboardInterrupt
        cls.kbi_event.set()
        cls.cls_log(logging.WARNING, 'Set the KbI Event after catching '
                                     'KeyboardInterrupt in the main Process '
                                     'in renice_all')
      # General case
      else:
        cls.logger.exception("Caught exception while running renice_all, "
                             "aborting", exc_info=exc)
        # Any Exception caught in the main Process must stop the script
        cls.raise_event.set()
        cls.cls_log(logging.WARNING, 'Set the raise Event after exception was '
                                     'caught in the main Process in '
                                     'renice_all')
      # Breaking the Barrier to warn other Processes that something went wrong
      cls.ready_barrier.abort()
      cls.cls_log(logging.WARNING, "Broke the Barrier due to an exception "
                                   "caught in renice_all")
      # Need to clean up the running Blocks and other Processes / Threads
      cls._cleanup()

  @classmethod
  def launch_all(cls, no_raise: bool = False) -> None:
    """The final method being called by the main
    :obj:`~multiprocessing.Process` running a script with Crappy.

    It unlocks all the Blocks by releasing the synchronization
    :obj:`~multiprocessing.Barrier`, sets the shared t0
    :obj:`~multiprocessing.Value`, and then waits for all the Blocks to finish.

    In case an exception is raised, sets the stop :obj:`~multiprocessing.Event`
    for warning the Blocks, waits for the Blocks to finish, and if they don't,
    terminates them.

    Args:
      no_raise: When set to :obj:`False`, the Exceptions encountered during
        Crappy's execution, as well as the :exc:`KeyboardInterrupt`, will raise
        an Exception right before Crappy returns. This is meant to prevent the
        execution of code that would come after Crappy, in case Crappy does not
        terminate as expected. This behavior can be disabled by setting this
        argument to :obj:`True`.
    
    .. versionremoved:: 2.0.0 *t0*, *verbose* and *bg* arguments
    """

    # Setting the no_raise flag
    cls.no_raise = no_raise

    # Flag indicating whether to perform the cleanup action or not
    cleanup = True

    try:
      # Making sure that the Block classmethods are called in the right order
      if not cls.prepared_all:
        cls.cls_log(logging.ERROR, "Cannot call launch_all before calling "
                                   "prepare_all ! Aborting")
        # If prepare wasn't called, there is no need to clean up Crappy
        cleanup = False
        raise RuntimeError("Cannot call launch before calling prepare ! "
                           "Aborting")

      if cls.launched_all:
        cls.cls_log(logging.ERROR, "The launched_all flag is unexpectedly "
                                   "raised, aborting !")
        raise RuntimeError

      cls.launched_all = True

      # The Barrier waits for the main Process to be ready so that the
      # prepare_all and launch_all methods can be used separately for a finer
      # grained control
      cls.cls_log(logging.INFO, 'Waiting for all Blocks to be ready')
      cls.ready_barrier.wait()
      cls.cls_log(logging.INFO, 'All Blocks ready now')

      # Setting t0 and telling all the Blocks to start
      cls.shared_t0.value = time_ns() / 1e9
      cls.cls_log(logging.INFO, f'Start time set to {cls.shared_t0.value}s')
      cls.start_event.set()
      cls.cls_log(logging.INFO, 'Start event set, all Blocks can now start')

      # The main Process mustn't finish before all the Blocks are stopped
      cls.cls_log(logging.INFO, 'Main Process done, waiting for all Blocks to '
                                'finish')
      for _ in wait([inst.sentinel for inst in cls.instances]):
        cls.cls_log(logging.INFO, "A Block has finished, waiting for the "
                                  "other ones to follow")

    except (BrokenBarrierError, KeyboardInterrupt, Exception) as exc:

      # If there is no specific cleanup to perform, only raising
      if not cleanup:
        raise

      # KeyboardInterrupt is a separate case
      if isinstance(exc, KeyboardInterrupt):
        cls.cls_log(logging.WARNING, "Caught KeyboardInterrupt in the main "
                                     "Process while running launch_all !")
        # Special Event for the KeyboardInterrupt
        cls.kbi_event.set()
        cls.cls_log(logging.WARNING, 'Set the KbI Event after catching '
                                     'KeyboardInterrupt in the main Process '
                                     'in launch_all')
      # Case when a Block crashed while preparing
      elif isinstance(exc, BrokenBarrierError):
        cls.cls_log(logging.ERROR, "Exception raised in a Block while waiting "
                                   "for all Blocks to be ready, stopping")
      # General case
      else:
        cls.logger.exception("Caught exception while running launch_all, "
                             "aborting", exc_info=exc)
        # Any Exception caught in the main Process must stop the script
        cls.raise_event.set()
        cls.cls_log(logging.WARNING, 'Set the raise Event after exception was '
                                     'caught in the main Process in '
                                     'launch_all')
      # Breaking the Barrier to warn other Processes that something went wrong
      cls.ready_barrier.abort()
      cls.cls_log(logging.WARNING, "Broke the Barrier due to an exception "
                                   "caught in launch_all")
    finally:
      # Need to clean up the running Blocks and other Processes / Threads
      if cleanup:
        cls._cleanup()

  @classmethod
  def _cleanup(cls) -> None:
    """Method called at the very end of every script execution.

    It first waits for all the Blocks to end, and kills them if they don't stop
    by themselves. Then, it also stops, if relevant, the USBServer and the
    log_thread, and warns the user in case Processes would still be running.

    Finally, it raises an exception if needed, in order to stop the script of
    the main Process. This way, any action that could follow the normal
    execution of Crappy won't happen, unless the user explicitly catches
    Crappy's exception and decides to go on with the script.
    """

    try:

      # Setting the stop Event, to indicate all the Blocks to finish
      cls.stop_event.set()
      cls.cls_log(logging.INFO, 'Stop event set, waiting for all Blocks to '
                                'finish')
      t = time()

      # Waiting at most 3 seconds for all the Blocks to finish
      while cls.instances and not all(not inst.is_alive() for inst
                                      in cls.instances):
        cls.cls_log(logging.INFO, "All Blocks not stopped yet")
        sleep(0.5)

        # After 3 seconds, killing the Blocks that didn't stop
        if time() - t > 3:
          cls.cls_log(logging.WARNING, 'All Blocks not stopped, terminating '
                                       'the living ones')
          for inst in cls.instances:
            if inst.is_alive():
              inst.terminate()
              cls.cls_log(logging.WARNING, f'Block {inst.name} terminated')
            else:
              cls.cls_log(logging.INFO, f'Block {inst.name} done')

          break

      # Stopping the USB server if required
      if USBServer.initialized:
        cls.cls_log(logging.INFO, "Stopping the USB server")
        USBServer.stop_server()

      # Stopping the log thread if required
      if get_start_method() == 'spawn' and cls.log_thread is not None:
        cls.thread_stop = True
        cls.log_thread.join(timeout=0.1)

      # Warning in case the log thread did not stop correctly
      if cls.log_thread is not None and cls.log_thread.is_alive():
        cls.cls_log(logging.WARNING, "The Thread reading the log messages did "
                                     "not terminate in time !")

      # Checking whether all Blocks terminated gracefully
      if cls.instances and any(inst.is_alive() for inst in cls.instances):
        running = ', '.join(inst.name for inst in cls.instances
                            if inst.is_alive())
        cls.cls_log(logging.ERROR, f"Crappy failed to finish gracefully, "
                                   f"Block(s) {running} still running !")
        # An Exception is raised in case all the Blocks don't finish gracefully
        cls.raise_event.set()
        cls.cls_log(logging.WARNING, 'Set the raise Event because all the '
                                     'Blocks did not terminate as requested')
      else:
        cls.cls_log(logging.INFO, 'All Blocks done, Crappy terminated '
                                  'gracefully !\n')

    # Exceptions at that point cannot really be handled, but should still raise
    # in the main Process
    except (Exception, KeyboardInterrupt) as exc:
      # KeyboardInterrupt is a separate case
      if isinstance(exc, KeyboardInterrupt):
        cls.cls_log(logging.WARNING, "Caught KeyboardInterrupt while "
                                     "cleaning up, ignoring it !")
        # Special Event for the KeyboardInterrupt
        cls.kbi_event.set()
        cls.cls_log(logging.WARNING, 'Set the KbI Event after catching '
                                     'KeyboardInterrupt while cleaning up')
      else:
        cls.logger.exception("Caught exception while cleaning up !",
                             exc_info=exc)

        # Any Exception caught in the main Process must stop the script
        cls.raise_event.set()
        cls.cls_log(logging.WARNING, 'Set the raise Event after exception was '
                                     'caught in the main Process while '
                                     'cleaning up')

    # Deciding whether to raise and stop the main Process, and also resetting
    finally:
      # The try/finally is needed to reset Crappy before the exception is
      # raised but after the class Events are accessed
      try:
        # Deciding whether to raise or not
        if cls.raise_event.is_set() and not cls.no_raise:
          cls.cls_log(logging.ERROR, "An error occurred during Crappy's "
                                     "execution, raising CrappyFail !")
          raise CrappyFail
        elif cls.kbi_event.is_set() and not cls.no_raise:
          cls.cls_log(logging.ERROR, "KeyboardInterrupt called while running "
                                     "Crappy, raising it !")
          raise KeyboardInterrupt("Crappy was stopped using CTRL+C ! To "
                                  "disable this Exception, set the no_raise "
                                  "argument of crappy.start() or "
                                  "crappy.launch() to True.")
      finally:
        # Always resetting Crappy as the Blocks and synchronization objects
        # won't come in use anymore
        cls.reset()

  @classmethod
  def _set_logger(cls) -> None:
    """Initializes the logging for the main Process.

    It creates two Stream Loggers, one for the info and debug levels displaying
    on stdout and one for the other levels displaying on stderr. It also
    creates a File Logger for saving the log to a log file.

    The levels WARNING and above are always being displayed in the terminal, no
    matter what the user chooses. Similarly, the INFO log level and above are
    always being saved to the log file.
    """

    # The Logger handling all messages
    crappy_log = logging.getLogger('crappy')

    if cls.log_level is not None:
      crappy_log.setLevel(cls.log_level)
    else:
      logging.disable()

    # In case there's no logging, no need to configure the handlers
    if cls.log_level is not None:
      # The two handlers for displaying messages in the console
      stream_handler = logging.StreamHandler(stream=stdout)
      stream_handler_err = logging.StreamHandler(stream=stderr)

      # Getting the path to Crappy's temporary folder
      if system() in ('Linux', 'Darwin'):
        log_path = Path('/tmp/crappy')
      elif system() == 'Windows':
        log_path = Path.home() / 'AppData' / 'Local' / 'Temp' / 'crappy'
      else:
        log_path = None

      # Creating Crappy's temporary folder if needed
      if log_path is not None:
        try:
          log_path.mkdir(parents=False, exist_ok=True)
        except FileNotFoundError:
          log_path = None

      # This handler writes the log messages to a log file
      if log_path is not None:
        file_handler = logging.FileHandler(log_path / 'logs.txt', mode='w')
      else:
        file_handler = None

      # Setting the log levels for the handlers
      stream_handler.setLevel(max(logging.DEBUG, cls.log_level))
      stream_handler.addFilter(cls._stdout_filter)
      stream_handler_err.setLevel(max(logging.WARNING, cls.log_level))
      if file_handler is not None:
        file_handler.setLevel(max(logging.DEBUG, cls.log_level))

      # Setting the log format for the handlers
      log_format = logging.Formatter('%(asctime)s %(name)s %(levelname)-8s '
                                     '%(message)s')
      stream_handler.setFormatter(log_format)
      stream_handler_err.setFormatter(log_format)
      if file_handler is not None:
        file_handler.setFormatter(log_format)

      # Adding the handlers to the Logger
      crappy_log.addHandler(stream_handler)
      crappy_log.addHandler(stream_handler_err)
      if file_handler is not None:
        crappy_log.addHandler(file_handler)

    cls.logger = crappy_log

  @classmethod
  def stop_all(cls) -> None:
    """Method for stopping all the Blocks by setting the stop
    :obj:`~multiprocessing.Event`.
    
    .. versionremoved:: 2.0.0 *verbose* argument
    """

    if cls.stop_event is not None:
      cls.stop_event.set()
      cls.cls_log(logging.INFO, 'Stop event set after a call to stop(), all '
                                'Blocks should now finish')

  @classmethod
  def reset(cls) -> None:
    """Resets Crappy by emptying the :obj:`~weakref.WeakSet` containing
    references to all the Blocks and resetting the synchronization objects.

    This method is called at the very end of the :meth:`_cleanup` method, but
    can also be called to "revert" the instantiation of Blocks while Crappy
    isn't started yet.
    """

    cls.instances = WeakSet()
    cls.names = list()
    cls.thread_stop = False
    cls.prepared_all = False
    cls.launched_all = False
    cls.no_raise = False

    cls.shared_t0 = None
    cls.ready_barrier = None
    cls.start_event = None
    cls.stop_event = None
    cls.raise_event = None
    cls.kbi_event = None

    if cls.logger is not None:
      cls.cls_log(logging.INFO, 'Crappy was successfully reset')
  
  @classmethod
  def cls_log(cls, level: int, msg: str) -> None:
    """Wrapper for logging messages in the main Process.
    
    Ensures the Logger exists before trying to log, thus avoiding potential 
    errors.
    
    .. versionadded:: 2.0.0
    """
    
    if cls.logger is None:
      return
    cls.logger.log(level=level, msg=msg)

  @classmethod
  def _log_target(cls) -> None:
    """This method is the target to the Logger Thread.

    It reads log messages from a Queue and passes them to the Logger for
    handling.
    """

    while not cls.thread_stop:
      try:
        record = cls.log_queue.get(block=True, timeout=0.05)
      except Empty:
        continue

      logger = logging.getLogger(record.name)
      logger.handle(record)

  @staticmethod
  def _stdout_filter(rec: logging.LogRecord) -> bool:
    """Returns :obj:`True` if the input log message has level INFO or DEBUG,
    :obj:`False` otherwise."""

    return rec.levelno in (logging.DEBUG, logging.INFO)

  def run(self) -> None:
    """The method run by the Blocks when their :obj:`~multiprocessing.Process` 
    is started.

    It first calls :meth:`prepare`, then waits at the
    :obj:`~multiprocessing.Barrier` for all Blocks to be ready, then calls
    :meth:`begin`, then :meth:`main`, and finally :meth:`finish`.
    
    If an exception is raised, sets the shared stop 
    :obj:`~multiprocessing.Event` to warn all the other Blocks.
    """

    try:
      # Any Exception caught at the beginning should break the Barrier
      try:
        # Initializes the Logger for the Block
        self._set_block_logger()
        self.log(logging.INFO, "Block launched")

        # Running the preliminary actions before the test starts
        self.log(logging.INFO, "Block preparing")
        self.prepare()

      # If an Exception is raised, warning the other Blocks by breaking the
      # Barrier
      except (Exception, KeyboardInterrupt):
        self._ready_barrier.abort()
        self.log(logging.WARNING, "Broke the Barrier after an Exception was "
                                  "caught while preparing")
        raise

      # Waiting for all Blocks to be ready, except if the Barrier was broken
      try:
        self.log(logging.INFO, "Waiting for the other Blocks to be ready")
        self._ready_barrier.wait()
        self.log(logging.INFO, "All Blocks ready now")
      except BrokenBarrierError:
        raise PrepareError

      # Waiting for t0 to be set, should take a few milliseconds at most
      self.log(logging.INFO, "Waiting for the start time to be set")
      self._start_event.wait(timeout=1)
      if not self._start_event.is_set():
        raise StartTimeout
      else:
        self.log(logging.INFO, "Start time set, Block starting")

      # Running the first loop
      self.log(logging.INFO, "Calling begin method")
      self.begin()

      # Setting the attributes for counting the performance
      self._last_t = time_ns() / 1e9
      self._last_fps = self._last_t
      self._n_loops = 0

      # Running the main loop until told to stop
      self.log(logging.INFO, "Entering main loop")
      self.main()
      self.log(logging.INFO, "Exiting main loop after stop Event was set")

    # A wrong data type was sent through a Link
    except LinkDataError:
      self.log(logging.ERROR, "Tried to send a wrong data type through a Link,"
                              " stopping !")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
    # An error occurred in another Block while preparing
    except PrepareError:
      self.log(logging.ERROR, "Exception raised in another Block while waiting"
                              " for all Blocks to be ready, stopping")
    # An error occurred in the CameraConfig window while preparing
    except CameraConfigError:
      self.log(logging.ERROR, "Exception raised in a configuration window, "
                              "stopping")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
    # An error occurred in a Camera process while preparing
    except CameraPrepareError:
      self.log(logging.ERROR, "Exception raised in a Camera Process while "
                              "preparing, stopping")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
      # An error occurred in a Camera Process while running
    except CameraRuntimeError:
      self.log(logging.ERROR, "Exception raised in a Camera process while "
                              "running, stopping")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
    # The start Event took too long to be set
    except StartTimeout:
      self.log(logging.ERROR, "Waited too long for start time to be set, "
                              "aborting !")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
    # Tried to access t0 but it's not set yet
    except T0NotSetError:
      self.log(logging.ERROR, "Trying to get the value of t0 when it's not "
                              "set yet, aborting")
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')
    # A Generator Block finished its Path
    except GeneratorStop:
      self.log(logging.WARNING, f"Generator Path exhausted, stopping the "
                                f"Block")
    # A FileReader Camera object has no more file to read from
    except ReaderStop:
      self.log(logging.WARNING, "Exhausted all the images to read from a "
                                "FileReader Camera, stopping the Block")
    # The user requested the script to stop
    except KeyboardInterrupt:
      self.log(logging.WARNING, f"KeyboardInterrupt caught, stopping")
      # A KeyboardInterrupt should stop the script and be raised as is
      self._kbi_event.set()
      self.log(logging.WARNING, 'Set the KbI Event after catching a '
                                'KeyboardInterrupt while running')
    # Another Exception occurred
    except (Exception,) as exc:
      self._logger.exception("Caught Exception while running !", exc_info=exc)
      # Any unexpected Exception should stop the script
      self._raise_event.set()
      self.log(logging.WARNING, 'Set the raise Event after catching an '
                                'unexpected Exception while running')

    # In all cases, trying to properly close the Block
    finally:
      try:
        self.log(logging.INFO, "Setting the stop Event")
        self._stop_event.set()
        self.log(logging.INFO, "Calling the finish method")
        self.finish()
      except KeyboardInterrupt:
        self.log(logging.WARNING, "Caught KeyboardInterrupt while finishing, "
                                  "ignoring it")
        # A KeyboardInterrupt should stop the script and be raised as is
        self._kbi_event.set()
        self.log(logging.WARNING, 'Set the KbI Event after catching a '
                                  'KeyboardInterrupt while finishing')
      except (Exception,) as exc:
        self._logger.exception("Caught Exception while finishing !",
                               exc_info=exc)
        # Any unexpected Exception should stop the script
        self._raise_event.set()
        self.log(logging.WARNING, 'Set the raise Event after catching an '
                                  'unexpected Exception while finishing')

  def main(self) -> None:
    """The main loop of the :meth:`run` method. Repeatedly calls the
    :meth:`loop` method and manages the looping frequency."""

    # Looping until told to stop or an error occurs
    while not self._stop_event.is_set():
      self.log(logging.DEBUG, "Looping")
      self.loop()
      self.log(logging.DEBUG, "Handling freq")
      self._handle_freq()

  def prepare(self) -> None:
    """This method should perform any action required for initializing the
    Block before the test starts.

    For example, it can open a network connection, create a file, etc. It is
    also fine for this method not to be overriden if there's no particular
    action to perform.

    Note that this method is called once the :obj:`~multiprocessing.Process`
    associated to the Block has been started.
    """

    ...

  def begin(self) -> None:
    """This method can be considered as the first loop of the test, and is
    called before the :meth:`loop` method.

    It allows to perform initialization actions that cannot be achieved in the
    :meth:`prepare` method.
    """

    ...

  def loop(self) -> None:
    """This method is the core of the Block. It is called repeatedly during the
    test, until the test stops or an error occurs.

    Only in this method should data be sent to downstream Blocks, or received
    from upstream Blocks.

    Although it is possible not to override this method, that has no practical
    interest and this method should always be rewritten.
    """

    self.log(logging.WARNING, f"Loop method not defined, this block does "
                              f"nothing !")
    sleep(1)

  def finish(self) -> None:
    """This method should perform any action required for properly ending the
    test.

    For example, it can close a file or disconnect from a network. It is also
    fine for this method not to be overriden if no particular action needs to
    be performed.

    Note that this method will normally be called even in case an error occurs,
    although that cannot be guaranteed.
    """

    ...

  def stop(self) -> None:
    """This method stops all the running Blocks.

    It should be called from the :meth:`loop` method of a Block. It allows to
    stop the execution of the script in a clean way, without raising an
    exception. It is mostly intended for users writing their own Blocks.

    Note:
      Calling this method in :meth:`__init__`, :meth:`prepare` or :meth:`begin`
      is not recommended, as the Block will only stop when reaching the
      :meth:`loop` method. Calling this method during :meth:`finish` will have
      no effect.
    """

    if self._stop_event is not None:
      self.log(logging.WARNING, "stop method called, setting the stop event !")
      self._stop_event.set()

  def _handle_freq(self) -> None:
    """This method ensures that the Block loops at the desired frequency, or as
    fast as possible if the requested frequency cannot be achieved.

    It also displays the looping frequency of the Block if requested by the
    user. If no looping frequency is specified, the Block will loop as fast as
    possible.
    """

    self._n_loops += 1
    t = time_ns() / 1e9

    # Only handling frequency if requested
    if self.freq is not None:

      # Correcting the error of the sleep function through a recursive approach
      # The last 2 milliseconds are in free loop
      remaining = self._last_t + 1 / self.freq - t
      while remaining > 0:
        t = time_ns() / 1e9
        remaining = self._last_t + 1 / self.freq - t
        sleep(max(0., remaining / 2 - 2e-3))

    self._last_t = t

    # Displaying frequency every 2 seconds
    if self.display_freq and self._last_t - self._last_fps > 2:
      self.log(
        logging.INFO,
        f"loops/s: {self._n_loops / (self._last_t - self._last_fps)}")

      self._n_loops = 0
      self._last_fps = self._last_t

  def _set_block_logger(self) -> None:
    """Initializes the Logger for the Block.

    If the :mod:`multiprocessing` start method is `spawn` (mostly on Windows),
    redirects the log messages to a Queue for passing them to the main Process.
    """

    logger = logging.getLogger(self.name)

    # Adjusting logging to the desired level
    if self._log_level is not None:
      logger.setLevel(self._log_level)
    else:
      logging.disable()

    # On Windows, the messages need to be sent through a Queue for logging
    if get_start_method() == "spawn" and self._log_level is not None:
      queue_handler = logging.handlers.QueueHandler(self._log_queue)
      queue_handler.setLevel(self._log_level)
      logger.addHandler(queue_handler)

    self._logger = logger

  @property
  def debug(self) -> Optional[bool]:
    """Indicates whether the debug information should be displayed or not.

    If :obj:`False` (the default), only displays the :obj:`~logging.INFO`
    logging level. If :obj:`True`, displays the :obj:`~logging.DEBUG` logging
    level for the Block. And if :obj:`None`, displays only the
    :obj:`~logging.CRITICAL` logging level, which is equivalent to no
    information at all.
    
    .. versionadded:: 2.0.0
    """

    return self._debug

  @debug.setter
  def debug(self, val: Optional[bool]) -> None:
    if val is not None:
      if val:
        self._debug = True
        self._log_level = logging.DEBUG
      else:
        self._debug = False
        self._log_level = logging.INFO
    else:
      self._debug = None
      self._log_level = logging.CRITICAL

  @property
  def t0(self) -> float:
    """Returns the value of t0, the exact moment when the test started that is
    shared between all the Blocks.
    
    .. versionadded:: 2.0.0
    """

    if self._instance_t0 is not None and self._instance_t0.value > 0:
      self.log(logging.DEBUG, "Start time value requested")
      return self._instance_t0.value
    else:
      raise T0NotSetError

  def add_output(self, link: Link) -> None:
    """Adds an output :class:`~crappy.links.Link` to the list of output Links
    of the Block."""

    self.outputs.append(link)

  def add_input(self, link: Link) -> None:
    """Adds an input :class:`~crappy.links.Link` to the list of input Links of
    the Block."""

    self.inputs.append(link)

  def log(self, log_level: int, msg: str) -> None:
    """Method for recording log messages from the Block. This option should be
    preferred to calling :func:`print`.

    Args:
      log_level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      return
    self._logger.log(log_level, msg)

  def send(self, data: Optional[Union[Dict[str, Any], Iterable[Any]]]) -> None:
    """Method for sending data to downstream Blocks.

    The exact same :obj:`dict` is sent to every downstream Block.

    This method accepts the data to send either as a :obj:`dict` or as another
    type of iterable (like a :obj:`list` or a :obj:`tuple`). If data is
    provided as a dict, it is sent as is. The keys of the dict then correspond
    to the labels. Otherwise, the values given as an iterable are first
    converted to a dict using the ``self.labels`` attribute containing the
    labels to use.

    It is up to the user to match the order of the values in the iterable with
    the order of the labels in ``self.labels``. If the number of labels and the
    number of values to send do not match, no error is raised but some data
    might not get sent.
    """

    # Just in case, not handling non-existing data
    if data is None:
      return

    # Case when the data to send is not given as a dict
    if not isinstance(data, dict):
      # First, checking that labels are provided
      if self.labels is None or not self.labels:
        self.log(logging.ERROR, "Trying to send data as an iterable, but no "
                                "labels are specified ! Please add a "
                                "self.labels attribute.")
        raise LinkDataError

      # Trying to convert iterable data to dict using the given labels
      try:
        self.log(logging.DEBUG, f"Converting {data} to dict before sending")
        data = dict(zip(self.labels, data))
      except TypeError:
        self.log(logging.ERROR, f"Cannot convert data to send (of type "
                                f"{type(data)}) to dict ! Please ensure that "
                                f"the data is given as an iterable, as well as"
                                f" self.labels.")
        raise

    # Sending the data to the downstream Blocks
    for link in self.outputs:
      self.log(logging.DEBUG, f"Sending {data} to Link {link.name}")
      link.send(data)

  def data_available(self) -> bool:
    """Returns :obj:`True` if there's data available for reading in at least
    one of the input :class:`~crappy.links.Link`.
    
    .. versionchanged:: 2.0.0 renamed from poll to data_available
    """

    self.log(logging.DEBUG, "Data availability requested")
    return self.inputs and any(link.poll() for link in self.inputs)

  def recv_data(self) -> Dict[str, Any]:
    """Reads the first available values from each incoming
    :class:`~crappy.links.Link` and returns them all in a single dict.

    The returned :obj:`dict` might not always have a fixed number of keys,
    depending on the availability of incoming data.

    Also, the returned values are the oldest available in the Links. See
    :meth:`recv_last_data` for getting the newest available values.

    Important:
      If data is received over a same label from different Links, part of it
      will be lost ! Always avoid using a same label twice in a Crappy script.

    Returns:
      A :obj:`dict` whose keys are the received labels and with a single value
      for each key (usually a :obj:`float` or a :obj:`str`).
    
    .. versionchanged:: 2.0.0 renamed from recv_all to recv_data
    """

    ret = dict()

    for link in self.inputs:
      ret.update(link.recv())

    self.log(logging.DEBUG, f"Called recv_data, got {ret}")
    return ret

  def recv_last_data(self, fill_missing: bool = True) -> Dict[str, Any]:
    """Reads all the available values from each incoming
    :class:`~crappy.links.Link`, and returns the newest ones in a single dict.

    The returned :obj:`dict` might not always have a fixed number of keys,
    depending on the availability of incoming data.

    Important:
      If data is received over a same label from different Links, part of it
      will be lost ! Always avoid using a same label twice in a Crappy script.

    Args:
      fill_missing: If :obj:`True`, fills up the missing data for the known
        labels. This way, the last value received from all known labels is
        always returned. It can of course not fill up missing data for labels
        that haven't been received yet.

    Returns:
      A :obj:`dict` whose keys are the received labels and with a single value
      for each key (usually a :obj:`float` or a :obj:`str`).

    .. versionremoved:: 1.5.10 *num* argument
    .. versionadded:: 1.5.10 *blocking* argument
    .. versionchanged:: 2.0.0 renamed from get_last to recv_last_data
    """

    # Initializing the buffer storing the last received values
    if self._last_values is None:
      self._last_values = [dict() for _ in self.inputs]

    ret = dict()

    # Storing the received values in the return dict and in the buffer
    for link, buffer in zip(self.inputs, self._last_values):
      data = link.recv_last()
      ret.update(data)
      buffer.update(data)

    # If requested, filling up the missing values in the return dict
    if fill_missing:
      for buffer in self._last_values:
        ret.update(buffer)

    self.log(logging.DEBUG, f"Called recv_last_data, got {ret}")
    return ret

  def recv_all_data(self,
                    delay: Optional[float] = None,
                    poll_delay: float = 0.1) -> Dict[str, List[Any]]:
    """Reads all the available values from each incoming
    :class:`~crappy.links.Link`, and returns them all in a single dict.

    The returned :obj:`dict` might not always have a fixed number of keys,
    depending on the availability of incoming data.

    Important:
      If data is received over a same label from different Links, part of it
      will be lost ! Always avoid using a same label twice in a Crappy script.
      See the :meth:`recv_all_data_raw` method for receiving data with no loss.

    Warning:
      As the time label is (normally) shared between all Blocks, the values
      returned for this label will be inconsistent and shouldn't be used !

    Args:
      delay: If given specifies a delay, as a :obj:`float`, during which the
        method acquired data before returning. All the data received during
        this delay is saved and returned. Otherwise, just reads all the
        available data and returns as soon as it is exhausted.
      poll_delay: If the ``delay`` argument is given, the Links will be polled
        once every this value seconds. It ensures that the method doesn't spam
        the CPU in vain.

    Returns:
      A :obj:`dict` whose keys are the received labels and with a :obj:`list`
      of received values for each key. The first item in the list is the oldest
      one available in the Link, the last item is the newest available.

    .. versionremoved:: 1.5.10 *num* argument
    .. versionadded:: 1.5.10 *blocking* argument
    .. versionchanged:: 2.0.0 renamed from get_all_last to recv_all_data
    """

    ret = defaultdict(list)
    t0 = time()

    # If simple recv_all, just receiving from all input links
    if delay is None:
      for link in self.inputs:
        ret.update(link.recv_chunk())

    # Otherwise, receiving during the given period
    else:
      while time() - t0 < delay:
        last_t = time()
        # Updating the list of received values
        for link in self.inputs:
          data = link.recv_chunk()
          for label, values in data.items():
            ret[label].extend(values)
        # Sleeping to avoid useless CPU usage
        sleep(max(0., last_t + poll_delay - time()))

    # Returning a dict, not a defaultdict
    self.log(logging.DEBUG, f"Called recv_all_data, got {dict(ret)}")
    return dict(ret)

  def recv_all_data_raw(self,
                        delay: Optional[float] = None,
                        poll_delay: float = 0.1) -> List[Dict[str, List[Any]]]:
    """Reads all the available values from each incoming
    :class:`~crappy.links.Link`, and returns them separately in a list of
    dicts.

    Unlike :meth:`recv_all_data` this method does not fuse the received data
    into a single :obj:`dict`, so it is guaranteed to return all the available
    data with no loss.

    Args:
      delay: If given specifies a delay, as a :obj:`float`, during which the
        method acquired data before returning. All the data received during
        this delay is saved and returned. Otherwise, just reads all the
        available data and returns as soon as it is exhausted.
      poll_delay: If the ``delay`` argument is given, the Links will be polled
        once every this value seconds. It ensures that the method doesn't spam
        the CPU in vain.

    Returns:
      A :obj:`list` containing :obj:`dict`, whose keys are the received labels
      and with a :obj:`list` of received value for each key.
    
    .. versionadded:: 2.0.0
    """

    ret = [defaultdict(list) for _ in self.inputs]
    t0 = time()

    # If simple recv_all, just receiving from all input links
    if delay is None:
      for dic, link in zip(ret, self.inputs):
        dic.update(link.recv_chunk())

    # Otherwise, receiving during the given period
    else:
      while time() - t0 < delay:
        last_t = time()
        # Updating the list of received values
        for dic, link in zip(ret, self.inputs):
          data = link.recv_chunk()
          for label, values in data.items():
            dic[label].extend(values)
        # Sleeping to avoid useless CPU usage
        sleep(max(0., last_t + poll_delay - time()))

    self.log(logging.DEBUG, f"Called recv_all_data_raw, got "
                            f"{[dict(dic) for dic in ret]}")
    return [dict(dic) for dic in ret]



# coding: utf-8

from typing import Optional, Dict, Any
import logging
from multiprocessing import current_process

from .meta_modifier import MetaModifier


class Modifier(metaclass=MetaModifier):
  """The base class for all Modifier classes, simply allowing to keep track of
  them.

  The Modifiers allow altering data from an input :class:`~crappy.blocks.Block`
  before it gets sent to an output Block. Each Modifier is associated to a
  :class:`~crappy.links.Link` linking the two Blocks. It is passed as an
  argument of the :meth:`~crappy.link` method instantiating the Link.

  It is preferable for every Modifier to be a child of this class, although
  that is not mandatory. A Modifier only needs to be a callable, i.e. a class
  defining the :meth:`__call__` method or a function.

  .. versionadded:: 1.4.0
  """

  def __init__(self, *_, **__) -> None:
    """Sets the logger attribute.

    .. versionchanged:: 2.0.0 now accepts args and kwargs
    """

    self._logger: Optional[logging.Logger] = None

  def __call__(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """The main method altering the inout data and returning the altered data.

    It should take a :obj:`dict` as its only argument, and return another
    :obj:`dict`. Both dicts should have their keys as :obj:`str`, representing
    the labels. Their values constitute the data flowing through the
    :class:`~crappy.links.Link`.

    Args:
      data: The data from the input :class:`~crappy.blocks.Block`, as a
        :obj:`dict`.

    Returns:
      Data to send to the output :class:`~crappy.blocks.Block`, as a
      :obj:`dict`. It is also fine for this method to return :obj:`None`, in
      which case no message is transmitted to the output Block.

    .. versionadded:: 2.0.0
    """

    self.log(logging.DEBUG, f"Received {data}")
    self.log(logging.WARNING, "The __call__ method is not defined, not "
                              "altering the data !")
    self.log(logging.DEBUG, f"Sending {data}")
    return data

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the Modifiers.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)



# coding: utf-8

from typing import Optional, Dict, Any, Union, Iterable
import logging

from .meta_modifier import Modifier


class TrigOnValue(Modifier):
  """Modifier passing the data to the downstream Block only if the value
  carried by a given label matches a given set of accepted values.

  Mostly useful to trigger Blocks in predefined situations.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Trig_on_value to TrigOnValue
  """

  def __init__(self,
               label: str,
               values: Union[Any, Iterable[Any]]) -> None:
    """Sets the args and initializes the parent class.

    Args:
      label: The name of the label to monitor.
      values: The values of ``label`` for which the data will be transmitted.
        Can be a single value, or an iterable of values (like a :obj:`list` or
        a :obj:`tuple`).

    .. versionchanged:: 1.5.10 renamed *name* argument to *label*
    """

    super().__init__()

    self._label = label

    if isinstance(values, str):
      values = (values,)
    else:
      try:
        iter(values)
      except TypeError:
        values = (values,)

    self._values = tuple(values)

  def __call__(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Checks if the value of ``label`` is in the predefined set of accepted
    values, and if so transmits the data.
    
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    if data[self._label] in self._values:
      self.log(logging.DEBUG, f"Sending {data}")
      return data

    self.log(logging.DEBUG, "Not returning any data")



# coding: utf-8

import logging
from ..actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Biaxe(Actuator):
  """This class can drive a Kollmorgen ServoStar 300 conditioner in speed.

  It is used at the LaMcube for driving a bi-axial tensile test machine, hence
  its name. The :class:`~crappy.actuator.ServoStar300` Actuator can drive the
  same hardware, but only in position.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 38400,
               timeout: float = 1) -> None:
    """Sets the instance attributes.

    Args:
      port: Path to the serial port to use for communication, e.g
        `'/dev/ttyS4'`.
      baudrate: The baud rate for the serial communication.
      timeout: The timeout for the serial communication.
    """

    self._ser = None

    super().__init__()

    self._port = port
    self._baudrate = baudrate
    self._timeout = timeout

    self._speed = None

  def open(self) -> None:
    """Opens the serial connection to the ServoStar."""

    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate {self._baudrate}")
    self._ser = serial.Serial(self._port, self._baudrate,
                              serial.EIGHTBITS, serial.PARITY_EVEN,
                              serial.STOPBITS_ONE, self._timeout)
    self._clear_errors()

  def set_speed(self, speed: float) -> None:
    """Sets a displacement speed for the motor. A value of `1` corresponds to
    `0.002 mm/s`"""

    speed = int(speed / 0.002)

    if speed != self._speed:
      self.log(logging.DEBUG, f"Writing b'J{ {speed}}\\r\\n' on port "
                              f"{self._port}")
      self._ser.write(f'J {speed}\r\n'.encode('ASCII'))
      self._speed = speed

  def stop(self) -> None:
    """Sets the speed of the motor to `0`.
    
    .. versionadded:: 2.0.0
    """

    if self._ser is not None:
      self.set_speed(0)

  def close(self) -> None:
    """Closes the serial connection to the ServoStar."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()

  def _clear_errors(self) -> None:
    """Clears the errors on the ServoStar."""

    self.log(logging.DEBUG, f"Writing b'CLRFAULT\\r\\n' on port {self._port}")
    self._ser.write(b"CLRFAULT\r\n")
    self.log(logging.DEBUG, f"Writing b'OPMODE 0\\r\\n EN\\r\\n' on port "
                            f"{self._port}")
    self._ser.write(b"OPMODE 0\r\n EN\r\n")



# coding: utf-8

import numpy as np
from multiprocessing import Process, current_process, get_start_method
from multiprocessing.synchronize import Event
from multiprocessing.queues import Queue
import logging
import logging.handlers
from typing import Optional
from functools import partial
from time import sleep


class HistogramProcess(Process):
  """This class is a :obj:`multiprocessing.Process` taking an image as an input 
  via a :obj:`multiprocessing.Pipe`, and returning the histogram of that image 
  in another :obj:`~multiprocessing.Pipe`.

  It is used by the :class:`~crappy.tool.camera_config.CameraConfig` window and 
  its children to delegate and parallelize the calculation of the histogram. It
  allows to gain a few frames per second on the display in the configuration
  window.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self,
               stop_event: Event,
               processing_event: Event,
               img_in: Queue,
               img_out: Queue,
               log_level: Optional[int],
               log_queue: Queue) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      stop_event: An :obj:`multiprocessing.Event` signaling the
        :obj:`~multiprocessing.Process` when to stop running.
      processing_event: An :obj:`multiprocessing.Event` set by the
        :obj:`multiprocessing.Process` to indicate that it's currently
        processing an image. Avoids having images to process piling up.
      img_in: The :obj:`~multiprocessing.queues.Queue` through which
        the images to process are received.
      img_out: The :obj:`~multiprocessing.queues.Queue` through which
        the calculated histograms are sent back.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to 
        the main :obj:`~logging.Logger`, only used in Windows.
    """

    self._logger: Optional[logging.Logger] = None
    self._log_level = log_level
    self._log_queue = log_queue

    super().__init__(name=f"{current_process().name}.{type(self).__name__}")

    self._stop_event: Event = stop_event
    self._processing_event: Event = processing_event
    self._img_in: Queue = img_in
    self._img_out: Queue = img_out

  def run(self) -> None:
    """The main method being run by the HistogramProcess.

    It continuously receives images from the 
    :class:`~crappy.tool.camera_config.CameraConfig`, calculates their 
    histograms and returns them back as a nice image to integrate on the
    window.
    """

    try:
      self._processing_event.clear()

      # Initializing the variables
      img, auto_range, low_thresh, high_thresh = None, None, None, None

      # Looping until told to stop or an exception is raised
      while not self._stop_event.is_set():

        # Setting the processing event when busy processing an image
        if not self._img_in.empty():
          self._processing_event.set()
          # Receiving the image to process as well as additional parameters
          while not self._img_in.empty():
            (img, auto_range,
             low_thresh, high_thresh) = self._img_in.get_nowait()

          self.log(logging.DEBUG, "Received image from CameraConfig")

          # Calculating the histogram
          hist, _ = np.histogram(img, bins=np.arange(257))
          hist = np.repeat(hist / np.max(hist) * 80, 2)
          hist = np.repeat(hist[np.newaxis, :], 80, axis=0)

          # Making a nice image out of the calculated histogram
          out_img = np.fromfunction(partial(self._hist_func, histo=hist),
                                    shape=(80, 512))
          out_img = np.flip(out_img, axis=0).astype('uint8')

          # Adding vertical grey bars to indicate the limits of the auto range
          if auto_range:
            self.log(logging.DEBUG, "Drawing the line of the auto-range")
            out_img[:, round(2 * low_thresh)] = 127
            out_img[:, round(2 * high_thresh)] = 127

          # Sending back the histogram
          self._img_out.put_nowait(out_img)
          self._processing_event.clear()
          self.log(logging.DEBUG, "Sent the histogram back to the "
                                  "CameraConfig")

        # To avoid spamming the CPU in vain when idle
        else:
          sleep(0.001)

      self.log(logging.INFO, "Stop event set, stopping")

    except KeyboardInterrupt:
      self.log(logging.INFO, "Caught KeyboardInterrupt, stopping")
    except (Exception,) as exc:
      self._logger.exception("Caught Exception while running, stopping !",
                             exc_info=exc)
    finally:
      self.log(logging.INFO, "HistogramProcess finished")

  @staticmethod
  def _hist_func(x: np.ndarray,
                 _: np.ndarray,
                 histo: np.ndarray) -> np.ndarray:
    """Function passed to the :meth:`numpy.fromfunction` method for building 
    the histogram."""

    return np.where(x <= histo, 0, 255)

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the HistogramProcess.

    Also instantiates the :obj:`~logging.Logger` when logging the first
    message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    """

    if self._logger is None:
      self._set_logger()

    self._logger.log(level, msg)

  def _set_logger(self) -> None:
    """Instantiates and sets up the logger for the HistogramProcess."""

    logger = logging.getLogger(self.name)

    # Disabling logging if requested
    if self._log_level is not None:
      logger.setLevel(self._log_level)
    else:
      logging.disable()

    # On Windows, the messages need to be sent through a Queue for logging
    if get_start_method() == "spawn" and self._log_level is not None:
      queue_handler = logging.handlers.QueueHandler(self._log_queue)
      queue_handler.setLevel(self._log_level)
      logger.addHandler(queue_handler)

    self._logger = logger



# coding: utf-8

from typing import Tuple, List, Any
import numpy as np
from time import time
import logging
from  warnings import warn

from .meta_camera import Camera
from .._global import OptionalModule

try:
  import usb.util
  import usb.core

  Seek_therm_usb_req = {'Write': usb.util.CTRL_OUT |
                        usb.util.CTRL_TYPE_VENDOR |
                        usb.util.CTRL_RECIPIENT_INTERFACE,
                        'Read': usb.util.CTRL_IN |
                        usb.util.CTRL_TYPE_VENDOR |
                        usb.util.CTRL_RECIPIENT_INTERFACE,
                        'Read_img': usb.util.CTRL_IN |
                        usb.util.CTRL_TYPE_STANDARD |
                        usb.util.CTRL_RECIPIENT_INTERFACE}

except (ModuleNotFoundError, ImportError):
  usb = OptionalModule("usb")

Seek_thermal_pro_vendor = 0x289D
Seek_thermal_pro_product = 0x0011

Seek_thermal_pro_commands = {'Read chip id': 0x36,
                             'Start get image transfer': 0x53,
                             'Get operation mode': 0x3D,
                             'Get image processing mode': 0x3F,
                             'Get firmware info': 0x4E,
                             'Get factory settings': 0x58,
                             'Set operation mode': 0x3C,
                             'Set image processing mode': 0x3E,
                             'Set firmware info features': 0x55,
                             'Set factory settings features': 0x56}

Seek_thermal_pro_dimensions = {'Width': 320,
                               'Height': 240,
                               'Raw width': 342,
                               'Raw height': 260}


class SeekThermalPro(Camera):
  """Class for reading images from the Seek Thermal Pro infrared camera.

  The SeekThermalPro Camera is meant for reading images from a Seek
  Thermal Pro infrared camera. It communicates over USB, and gets images by
  converting the received bytearrays into :mod:`numpy` arrays.

  Important:
    **Only for Linux users:** In order to drive the Seek Thermal Pro, the
    appropriate udev rule should be set. This can be done using the
    `udev_rule_setter` utility in ``crappy``'s `util` folder. It is also
    possible to add it manually by running:
    ::

      $ echo "SUBSYSTEM==\\"usb\\", ATTR{idVendor}==\\"289d\\", \
MODE=\\"0777\\\"" | sudo tee seek_thermal.rules > /dev/null 2>&1

    in a shell opened in ``/etc/udev/rules.d``.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Seek_thermal_pro to SeekThermalPro
  """

  def __init__(self) -> None:
    """Selects the right USB device."""

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    super().__init__()

    self._dev = None
    self._calib = None
    self._dead_pixels = []

    # Listing all the matching USB devices
    devices = usb.core.find(find_all=True,
                            idVendor=Seek_thermal_pro_vendor,
                            idProduct=Seek_thermal_pro_product)
    devices = list(devices)

    # Making sure there's exactly one possible camera to read images from
    if len(devices) > 1:
      raise IOError("Several matching cameras found, impossible to "
                    "differentiate between them")
    elif len(devices) == 0:
      raise IOError("No matching camera found")
    else:
      self._dev = devices[0]

  def open(self) -> None:
    """Sets the USB communication and initializes the device."""

    # Setting the USB configuration on the camera
    try:
      self.log(logging.INFO, f"Setting configuration on USB device "
                             f"{self._dev}")
      self._dev.set_configuration()
    except usb.core.USBError:
      self.log(logging.ERROR,
               "An error occurred while setting the configuration of the USB"
               " device !\nYou may have to install the udev-rules for this "
               "USB device, this can be done using the udev_rule_setter "
               "utility in the util folder")
      raise

    # Initializing the camera by sending various commands to it
    self.log(logging.INFO, "Configuring the camera")
    self._write_data(Seek_thermal_pro_commands['Set operation mode'],
                     b'\x00\x00')
    self._write_data(
      Seek_thermal_pro_commands['Set factory settings features'],
      b'\x06\x00\x08\x00\x00\x00')
    self._write_data(Seek_thermal_pro_commands['Set firmware info features'],
                     b'\x17\x00')
    self._write_data(
      Seek_thermal_pro_commands['Set factory settings features'],
      b'\x01\x00\x00\x06\x00\x00')
    for i in range(10):
      for j in range(0, 256, 32):
        self._write_data(
          Seek_thermal_pro_commands['Set factory settings features'],
          b'\x20\x00' + bytes([j, i]) + b'\x00\x00')
    self._write_data(Seek_thermal_pro_commands['Set firmware info features'],
                     b'\x15\x00')
    self._write_data(Seek_thermal_pro_commands['Set image processing mode'],
                     b'\x08\x00')
    self._write_data(Seek_thermal_pro_commands['Set operation mode'],
                     b'\x01\x00')

    # Acquiring the dead pixels image and saving the dead pixes map
    self.log(logging.INFO, "Getting the dead pixels")
    for i in range(5):
      status, ret = self._grab()
      if status == 4:
        self._dead_pixels = self._get_dead_pixels_list(ret)
        break
      elif i == 4:
        self.log(logging.WARNING, "Could not get the dead pixels frame")

    # Acquiring the calibration image and calibrating the camera
    self.log(logging.INFO, "Calibrating the camera")
    for i in range(10):
      status, img = self._grab()
      if status == 1:
        self._calib = self._crop(img) - 1600
        break
      elif i == 9:
        raise TimeoutError("Could not set the camera")

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Reads a single image from the camera.

    Returns:
      The captured image as well as a timestamp.
    """

    count = 0
    # Looping until a valid frame is acquired
    while True:

      # Capturing one frame
      t = time()
      status, img = self._grab()

      # If a calibration frame is acquired, recalibrating
      if status == 1:
        self.log(logging.DEBUG, "Recalibrating the camera")
        self._calib = self._crop(img) - 1600

      # If a valid frame is acquired, returning it along with its metadata
      elif status == 3 and self._calib is not None:
        return t, self._correct_dead_pixels(self._crop(img)-self._calib)

      # If no valid image can be read, that's bad news
      elif count == 5:
        raise TimeoutError("Unable to read image")
      count += 1

  def close(self) -> None:
    """Resets the camera and releases the USB resources."""

    if self._dev is not None:
      for _ in range(3):
        self._write_data(Seek_thermal_pro_commands['Set operation mode'],
                         b'\x00\x00')
      self.log(logging.INFO, "Releasing the USB resources")
      usb.util.dispose_resources(self._dev)

  def _grab(self) -> [bytes, np.ndarray]:
    """Captures a raw image from the camera.

    Returns:
      The status information and the raw image.
    """

    # Sending the read command
    self._write_data(Seek_thermal_pro_commands['Start get image transfer'],
                     b'\x58\x5b\x01\x00')

    to_read = 2 * \
        Seek_thermal_pro_dimensions['Raw width'] * \
        Seek_thermal_pro_dimensions['Raw height']
    ret = bytearray()

    # Reading all the chunks containing the frame information
    while to_read - len(ret) > 512:
      ret += self._dev.read(
        endpoint=Seek_therm_usb_req['Read_img'],
        size_or_buffer=int(to_read / (Seek_thermal_pro_dimensions['Raw height']
                           / 20)),
        timeout=1000)

    # Returning the read image in the right format and the associated status
    status = ret[4]
    if len(ret) == to_read:
      return status, np.frombuffer(ret, dtype=np.uint16).reshape(
        Seek_thermal_pro_dimensions['Raw height'],
        Seek_thermal_pro_dimensions['Raw width'])
    else:
      return status, None

  def _get_dead_pixels_list(self, data: np.ndarray) -> List[Tuple[Any]]:
    """Identifies the dead pixels on an image.

    Args:
      data: The image to identify dead pixels on.

    Returns:
      A :obj:`list` containing the indexes of the dead pixels.
    """

    img = self._crop(np.frombuffer(data, dtype=np.uint16).reshape(
      Seek_thermal_pro_dimensions['Raw height'],
      Seek_thermal_pro_dimensions['Raw width']))
    return list(zip(*np.where(img < 100)))

  @staticmethod
  def _crop(raw_img: np.ndarray) -> np.ndarray:
    """Simply crops an image to the right dimensions."""

    return raw_img[4: 4 + Seek_thermal_pro_dimensions['Height'],
                   1: 1 + Seek_thermal_pro_dimensions['Width']]

  def _correct_dead_pixels(self, img: np.ndarray) -> np.ndarray:
    """Corrects the dead pixels values.

    The new value is the average value of the surrounding pixels.

    Args:
      img: The image to correct.

    Returns:
      The corrected image.
    """

    for i, j in self._dead_pixels:
      img[i, j] = np.median(img[max(0, i - 1): i + 2, max(0, j - 1): j + 2])
    return img

  def _write_data(self, request: int, data: bytes) -> int:
    """Wrapper for sending USB messages."""

    self.log(logging.DEBUG, f"Sending USB command with request type "
                            f"{Seek_therm_usb_req['Write']}, request "
                            f"{request}, value {0}, index {0},length or "
                            f"data {data}")

    try:
      return self._dev.ctrl_transfer(bmRequestType=Seek_therm_usb_req['Write'],
                                     bRequest=request,
                                     wValue=0,
                                     wIndex=0,
                                     data_or_wLength=data,
                                     timeout=None)
    except usb.core.USBError:
      raise IOError("An error occurred during USB communication")

  def _read_data(self, request: int, data: bytes) -> int:
    """Wrapper for reading USB messages."""

    self.log(logging.DEBUG, f"Sending USB command with request type "
                            f"{Seek_therm_usb_req['Read']}, request {request},"
                            f" value {0}, index {0},length or data {data}")

    try:
      return self._dev.ctrl_transfer(bmRequestType=Seek_therm_usb_req['Read'],
                                     bRequest=request,
                                     wValue=0,
                                     wIndex=0,
                                     data_or_wLength=data,
                                     timeout=None)
    except usb.core.USBError:
      raise IOError("An error occurred during USB communication")



# coding: utf-8

from time import sleep, time
from typing import Iterable, Optional
from re import fullmatch
import logging
from  warnings import warn

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from serial import Serial
  from serial.serialutil import SerialException
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule("pyserial")
  SerialException = OptionalModule("pyserial")


class Sim868(InOut):
  """This class can drive a SIM868 cellular module so that it sends SMS to
  given phone numbers.

  Important:
    This InOut should be associated with a :class:`~crappy.modifier.Modifier`
    to manage the messages to send.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Gsm to Sim868
  """

  def __init__(self,
               numbers: Iterable[str],
               port: str = "/dev/ttyUSB0",
               baudrate: int = 115200,
               pin_code: Optional[str] = None,
               registration_timeout: float = 10) -> None:
    """Checks the validity of the arguments.

    Args:
      numbers: An iterable (like a :obj:`list` or a :obj:`tuple`) of numbers
        the messages will be sent to. The syntax is the following :
        ::

          ["0611223344"]

      port: Serial port the Sim868 is connected to.
      baudrate: Serial baudrate, between `1200` and `115200`.
      pin_code: Optionally, a pin code to use for activating the SIM card.
      registration_timeout: The maximum number of seconds to allow for the
        Sim868 to register to a network once the SIM card has the ready status.
    
    .. versionadded:: 2.0.0 *pin_code* and *registration_timeout* arguments
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._ser = None

    super().__init__()

    self._port = port
    self._baudrate = baudrate
    self._pin = pin_code
    self._reg_timeout = registration_timeout

    # Change the type of numbers to bytes rather than string
    self._numbers = list(numbers)

  def open(self) -> None:
    """Initializes the Sim868 device and checks its network connection.

    First, the serial connection is checked. Then, checking if the SIM card
    requires a PIN code. If so and if one is given, sets the PIN code on the
    SIM. Then, checks that the SIM is connected to a network. Finally, sets the
    input mode to Text for the SMS.
    """

    try:
      self.log(logging.INFO, f"Opening the serial port {self._port} with "
                             f"baudrate {self._baudrate}")
      self._ser = Serial(self._port, self._baudrate, timeout=0)
    except SerialException:
      raise SerialException("Sim868 not connected or wrong port")

    # Polling the Sim868 and waiting for a response
    self.log(logging.DEBUG, f"Writing b'AT\\r' to port {self._port}")
    self._ser.write(b'AT\r')

    t = time()
    ret = ''
    while time() - t < 2:
      # Reading the answers
      ret = self._ser.readline().strip().decode()
      if ret:
        self.log(logging.DEBUG, f"Read {ret} from port {self._port}")
      # Just waiting for the answer to be 'OK'
      if 'OK' in ret:
        self.log(logging.INFO, "Successfully connected to Sim868")
        break
      sleep(0.1)

    # Raising if no response after 2 seconds
    if 'ERROR' in ret:
      raise IOError("Got an ERROR message from the Sim868")
    elif 'OK' not in ret:
      raise ConnectionError("Could not get answer from Sim868 after 2 seconds")

    # Checking if the status of the SIM card is OK
    self._ser.write(b'AT+CPIN?\r')
    self.log(logging.DEBUG, f"Writing b'AT+CPIN?\\r' to port {self._port}")

    # Waiting for a response from the Sim868
    t = time()
    ret = ''
    need_pin = False
    while time() - t < 2:
      # Reading the answers
      ret = self._ser.readline().strip().decode()
      if ret:
        self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

      # Expecting an answer giving the SIM status, and parsing it
      status = fullmatch(r'\+CPIN:\s(.+)', ret)
      if status is not None:
        status, = status.groups()
        # Several possible answers from the Sim868
        if status == 'READY':
          self.log(logging.INFO, "SIM card not requiring a pin code")
        elif status == 'SIM PIN':
          self.log(logging.INFO, "SIM card requiring a pin code")
          need_pin = True
        else:
          raise IOError(f"Got CPIN status from the Sim868: {status}, but the "
                        f"InOut does not implement this case")

      # Only exiting when receiving the final answer 'OK'
      if 'OK' in ret:
        self.log(logging.INFO, "Successfully read CPIN from Sim868")
        break
      sleep(0.1)

    # Raising if no response after 2 seconds
    if 'ERROR' in ret:
      raise IOError("Got an ERROR message from the Sim868")
    elif 'OK' not in ret:
      raise ConnectionError("Could not get answer from Sim868 after 2 seconds")

    # Case when a PIN code is needed
    if need_pin:
      # No PIN code was given
      if self._pin is None:
        raise ValueError("A PIN code is needed to activate the SIM card but "
                         "none was given")
      else:
        # Setting the PIN code for the SIM card
        self._ser.write(f'AT+CPIN={self._pin}\r'.encode())
        self.log(logging.DEBUG, f"Writing b'AT+CPIN={self._pin}\\r' to port "
                                f"{self._port}")

        # Waiting for a response from the Sim868
        t = time()
        ret = ''
        while time() - t < 2:
          # Reading the answers
          ret = self._ser.readline().strip().decode()
          if ret:
            self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

          # Only exiting when receiving the final answer 'OK'
          if 'OK' in ret:
            self.log(logging.INFO, "Successfully sent the PIN code")
            break
          sleep(0.1)

        # Raising if no response after 2 seconds
        if 'ERROR' in ret:
          raise IOError("Got an ERROR message from the Sim868")
        elif 'OK' not in ret:
          raise ConnectionError(
              "Could not get answer from Sim868 after 2 seconds")

        # Giving some time to the Sim868 for setting the PIN code
        sleep(1)

        # Checking again if the status of the SIM card is OK
        self._ser.write(b'AT+CPIN?\r')
        self.log(logging.DEBUG, f"Writing b'AT+CPIN?\\r' to port {self._port}")

        # Waiting for a response from the Sim868
        t = time()
        ret = ''
        while time() - t < 2:
          # Reading the answers
          ret = self._ser.readline().strip().decode()
          if ret:
            self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

          # Expecting an answer giving the SIM status, and parsing it
          status = fullmatch(r'\+CPIN:\s(.+)', ret)
          if status is not None:
            status, = status.groups()
            # Several possible answers from the Sim868
            if status == 'READY':
              self.log(logging.INFO, "Successfully set the PIN code")
            else:
              raise IOError("PIN still not ready after setting the PIN code")

          # Only exiting when receiving the final answer 'OK'
          if 'OK' in ret:
            self.log(logging.INFO, "Successfully read CPIN from Sim868")
            break
          sleep(0.1)

        # Raising if no response after 2 seconds
        if 'ERROR' in ret:
          raise IOError("Got an ERROR message from the Sim868")
        elif 'OK' not in ret:
          raise ConnectionError(
              "Could not get answer from Sim868 after 2 seconds")

    # Checking that the SIM card is registered with an operator
    registered = False
    t = time()
    while time() - t < self._reg_timeout:

      # Requesting the registration status from the Sim868
      self._ser.write(b'AT+CGREG?\r')
      self.log(logging.DEBUG, f"Writing b'AT+CGREG?\\r' to port {self._port}")

      ret = ''
      t1 = time()
      # Reading the answers
      while time() - t < self._reg_timeout:
        # Sometimes the Sim868 sends back non UTF-8 characters, ignoring them
        ret = self._ser.readline().strip().decode()
        if ret:
          self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

        # Expecting an answer giving the registration status, and parsing it
        status = fullmatch(r'\+CGREG:\s\d,(\d).*', ret)
        if status is not None:
          status, = status.groups()
          # Several possible answers from the Sim868
          if status in ('0', '3', '4'):
            raise IOError(f"The Sim868 is not connected to a network, and is "
                          f"not searching anymore")
          elif status == '2':
            self.log(logging.INFO, "The Sim868 is searching for an operator "
                                   "to register with")
          elif status in ('1', '5'):
            self.log(logging.INFO, "Sim868 successfully registered with an "
                                   "operator")
            registered = True

        # Only exiting when receiving the final answer 'OK'
        if 'OK' in ret:
          self.log(logging.DEBUG, "Successfully read registration status from "
                                  "Sim868")
          break
        sleep(0.1)

      # Raising if no response after the given timeout
      if 'ERROR' in ret:
        raise IOError("Got an ERROR message from the Sim868")
      elif 'OK' not in ret:
        raise ConnectionError(f"Could not get answer from Sim868 after "
                              f"{t + self._reg_timeout - t1} seconds")

      # Exiting the loop once the Sim868 is connected to a network
      if registered:
        self.log(logging.INFO, "Sim868 successfully registered with an "
                               "operator")
        break
      sleep(self._reg_timeout / 5)

    if not registered:
      raise ConnectionError("The Sim868 was not able to register with an "
                            "operator before the given timeout")

    # Setting the message input mode to Text
    self._ser.write(b'AT+CMGF=1\r')
    self.log(logging.DEBUG, f"Writing b'AT+CMGF=1\\r' to port {self._port}")

    # Waiting for a response from the Sim868
    t = time()
    ret = ''
    while time() - t < 9:
      # Reading the answers
      ret = self._ser.readline().strip().decode()
      if ret:
        self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

      # Only exiting when receiving the final answer 'OK'
      if 'OK' in ret:
        self.log(logging.INFO, "Successfully set the message input mode")
        break
      sleep(0.1)

    # Raising if no response after 9 seconds
    if 'ERROR' in ret:
      raise IOError("Got an ERROR message from the Sim868")
    elif 'OK' not in ret:
      raise ConnectionError("Could not get answer from Sim868 after 9 seconds")

  def set_cmd(self, *cmd: str) -> None:
    """Sends an SMS whose text is the :obj:`str` received as command to all the
    phone numbers.

    If multiple messages are received, sends them all in the received order. If
    the messages are not :obj:`str`, they are first converted to strings if
    possible.
    """

    # Converting the messages to string
    cmd = map(str, cmd)

    # Iterating over all the messages to send
    for msg in cmd:
      # Not sending if the message is empty
      if msg:
        # Iterating over all the destination numbers
        for nr in self._numbers:

          # Providing the number to send the message to
          self._ser.write(f'AT+CMGS="{nr}"\r'.encode())
          self.log(logging.DEBUG, f"Writing b'AT+CMGS=\"{nr}\"\\r' to port "
                                  f"{self._port}")

          # Waiting for a response from the Sim868
          t = time()
          ret = ''
          while time() - t < 2:
            # Reading the answers
            while self._ser.inWaiting() > 0:
              ret += self._ser.read().decode()
            if ret:
              self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

            # Only exiting when receiving the > character
            if '> ' in ret:
              self.log(logging.INFO, "Received stop character for providing "
                                     "the message")
              break
            sleep(0.1)

          # Raising if no stop character after 2 seconds
          if 'ERROR' in ret:
            raise IOError("Got an ERROR message from the Sim868")
          elif '> ' not in ret:
            raise ValueError("Did not receive the stop character in 2 seconds")

          # Providing the text message to send
          self._ser.write(f"{msg}\x1a".encode())
          self.log(logging.DEBUG, f"Writing b'{msg}\\x1a' to port "
                                  f"{self._port}")

          # Waiting for a response from the Sim868
          t = time()
          ret = ''
          while time() - t < 2:
            # Reading the answers
            ret = self._ser.readline().strip().decode()
            if ret:
              self.log(logging.DEBUG, f"Read {ret} from port {self._port}")

            # Only exiting when receiving the final answer 'OK'
            if 'OK' in ret:
              self.log(logging.INFO, "Successfully sent the message")
              break
            sleep(0.1)

          # Raising if no response after 2 seconds
          if 'ERROR' in ret:
            raise IOError("Got an ERROR message from the Sim868")
          elif 'OK' not in ret:
            raise ConnectionError(
              "Could not get answer from Sim868 after 2 seconds")

  def close(self) -> None:
    """Closes the serial port."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()



# coding: utf-8

from .box import Box
from .histogram_process import HistogramProcess
from .overlay_object import Overlay
from .spots_boxes import SpotsBoxes
from .spots_detector import SpotsDetector
from .zoom import Zoom



# coding: utf-8

from typing import Optional, Callable, Union, Tuple, Iterable
import numpy as np
from pathlib import Path

from .camera_processes import DICVEProcess
from .camera import Camera
from ..tool.camera_config import DICVEConfig, SpotsBoxes
from .._global import CameraConfigError


class DICVE(Camera):
  """This Block can perform video-extensometry on images acquired by a
  :class:`~crappy.camera.Camera` object, by tracking patches using Digital
  Image Correlation techniques.

  It takes no input :class:`~crappy.links.Link` in a majority of situations,
  and outputs the results of the video-extensometry. It is a subclass of the
  :class:`~crappy.blocks.Camera` Block, and inherits of all its features. That
  includes the possibility to record and to display images in real-time,
  simultaneously to the image acquisition and processing. Refer to the
  documentation of the Camera Block for more information on these features.

  This Block is quite similar to the :class:`~crappy.blocks.VideoExtenso`
  Block, except this latter tracks spots instead of patches with a texture.
  Both Blocks output similar information, although the default labels and the
  data format are slightly different. The :class:`~crappy.blocks.DISCorrel`
  Block also relies on image correlation techniques for estimating the
  strain and the displacement on acquired images, but it only performs
  correlation on a single patch and is designed to have a much greater
  accuracy on this single patch. The :class:`~crappy.blocks.GPUVE` Block also
  performs video-extensometry based on digital image correlation, but the
  correlation is GPU-accelerated. The algorithm used for the correlation is
  also different from the ones available in this Block.

  For tracking the provided patches, several image correlation techniques are
  available. The most effective one is DISFlow, for which many parameters can
  be tuned. The other techniques are lighter on the CPU but also less precise.
  For each image, several values are computed and sent to the downstream
  Blocks. See the ``labels`` argument for a complete list.

  Similar to the :class:`~crappy.tool.camera_config.CameraConfig` window that
  can be displayed by the Camera Block, this Block can display a
  :class:`~crappy.tool.camera_config.DICVEConfig` window before the test
  starts. Here, the user can also select the patches to track if they were not
  already specified as an argument.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from DISVE to DICVE
  """

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               display_freq: bool = False,
               freq: Optional[float] = 200,
               debug: Optional[bool] = False,
               save_images: bool = False,
               img_extension: str = "tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               img_shape: Optional[Tuple[int, int]] = None,
               img_dtype: Optional[str] = None,
               patches: Optional[Iterable[Tuple[int, int, int, int]]] = None,
               labels: Optional[Union[str, Iterable[str]]] = None,
               method: str = 'Disflow',
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3,
               border: float = 0.2,
               safe: bool = True,
               follow: bool = True,
               raise_on_patch_exit: bool = True,
               **kwargs) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      camera: The name of the :class:`~crappy.camera.Camera` object to use for
        acquiring the images. Arguments can be passed to this Camera as
        ``kwargs`` of this Block. This argument is ignored if the
        ``image_generator`` argument is provided.
      transform: A callable taking an image as an argument, and returning a
        transformed image as an output. Allows applying a post-processing
        operation to the acquired images. This is done right after the
        acquisition, so the original image is permanently lost and only the
        transformed image is displayed and/or saved and/or further processed.
        The transform operation is not parallelized, so it might negatively
        affect the acquisition framerate if it is too heavy.
      config: If :obj:`True`, a
        :class:`~crappy.tool.camera_config.DICVEConfig` window is displayed
        before the test starts. There, the user can interactively adjust the
        different
        :class:`~crappy.camera.meta_camera.camera_setting.CameraSetting`
        available for the selected :class:`~crappy.camera.Camera`, visualize
        the acquired images, and select the patches to track if they haven't
        been given in the ``patches`` argument. The test starts when closing
        the configuration window. If not enabled, the ``img_dtype``,
        ``img_shape`` and ``patches`` arguments must be provided.
      display_images: If :obj:`True`, displays the acquired images in a
        dedicated window, using the backend given in ``displayer_backend`` and
        at the frequency specified in ``displayer_framerate``. This option
        should be considered as a debug or basic follow-up feature, it is not
        intended to be very fast nor to display high-quality images. The
        maximum resolution of the displayed images in `640x480`, the images
        might be downscaled to fit in this format. In addition to the acquired
        frames, the tracked patches are also displayed on the image as an
        overlay.
      displayer_backend: The backend to use for displaying the images. Can be
        either ``'cv2'`` or ``'mpl'``, to use respectively :mod:`cv2` (OpenCV)
        or :mod:`matplotlib`. ``'cv2'`` usually allows achieving a higher
        display frequency. Ignored if ``display_images`` is :obj:`False`. If
        not given and ``display_images`` is :obj:`True`, ``'cv2'`` is tried
        first and ``'mpl'`` second, and the first available one is used.
      displayer_framerate: The maximum update frequency of the image displayer,
        as an :obj:`int`. This value usually lies between 5 and 30Hz, the
        default is 5. The achieved update frequency might be lower than
        requested. Ignored if ``display_images`` is :obj:`False`.
      software_trig_label: The name of a label used as a software trigger for
        the :class:`~crappy.camera.Camera`. If given, images will only be
        acquired when receiving data over this label. The received value does
        not matter. This software trigger is not meant to be very precise, it
        is recommended not to rely on it for a trigger frequency greater than
        10Hz, in which case a hardware trigger should be preferred if available
        on the camera.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      save_images: If :obj:`True`, the acquired images are saved to the folder
        specified in ``save_folder``, in the format specified in
        ``img_extension``, using the backend specified in ``save_backend``, and
        at the frequency specified in ``save_period``. Each image is saved with
        the name : ``<frame_nr>_<timestamp>.<extension>``, and can thus easily
        be identified. Along with the images, a ``metadata.csv`` file records
        the metadata of all the saved images. This metadata is either the one
        returned by the :meth:`~crappy.camera.Camera.get_image` method of the
        :class:`~crappy.camera.Camera` object, or the default one generated in
        the :meth:`~crappy.blocks.Camera.loop` method of the
        :class:`~crappy.blocks.Camera` Block. Depending on the framerate of the
        camera and the performance of the computer, it is not guaranteed that
        all the acquired images will be recorded.
      img_extension: The file extension for the recorded images, as a
        :obj:`str` and without the dot. Common file extensions include `tiff`,
        `png`, `jpg`, etc. Depending on the used ``save_backend``, some
        extensions might not be available. It is currently not possible to
        customize the save parameters further than choosing the file extension.
        Ignored if ``save_images`` is :obj:`False`.
      save_folder: Path to the folder where to save the images, either as a
        :obj:`str` or as a :obj:`pathlib.Path`. Can be an absolute or a
        relative path, pointing to a folder. If the folder does not exist, it
        will be created (if the user has permission). If the given folder
        already contains a ``metadata.csv`` file (and thus likely images from
        Crappy), images are saved to another folder with the same name except
        a suffix is appended. Ignored if ``save_images`` is :obj:`False`. If
        not provided and ``save_images`` is :obj:`True`, the images are saved
        to the folder ``Crappy_images``, created next to the running script.
      save_period: Must be given as an :obj:`int`. Only one out of that number
        images at most will be saved. Allows to have a known periodicity in
        case the framerate is too high to record all the images. Or simply to
        reduce the number of recorded images if saving them all is not needed.
        Ignored if ``save_images`` is :obj:`False`.
      save_backend: If ``save_images`` is :obj:`True`, the backend to use for
        recording the images. It should be one of:
        ::

          'sitk', 'cv2', 'pil', 'npy'

        They correspond to the modules :mod:`SimpleITK`, :mod:`cv2` (OpenCV),
        :mod:`PIL` (Pillow Fork), and :mod:`numpy`. Note that the ``'npy'``
        backend saves the images as raw :obj:`numpy.array`, and thus ignores
        the ``img_extension`` argument. Depending on the machine, some backends
        may be faster or slower. For using each backend, the corresponding
        Python must of course be installed. If not provided and ``save_images``
        is :obj:`True`, the backends are tried in the same order as given above
        and the first available one is used. ``'npy'`` is always available.
      image_generator: A callable taking two :obj:`float` as arguments and
        returning an image as a :obj:`numpy.array`. **This argument is intended
        for use in the examples of Crappy, to apply an artificial strain on a
        base image. Most users should ignore it.** When given, the ``camera``
        argument is ignored and the images are acquired from the generator. To
        apply a strain on the image, strain values (in `%`) should be sent to
        the Camera Block over the labels ``'Exx(%)'`` and ``'Eyy(%)'``.
      img_shape: The shape of the images returned by the
        :class:`~crappy.camera.Camera` object as a :obj:`tuple` of :obj:`int`.
        It should correspond to the value returned by :obj:`numpy.shape`.
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      img_dtype: The `dtype` of the images returned by the
        :class:`~crappy.camera.Camera` object, as a :obj:`str`. It should
        correspond to a valid data type in :mod:`numpy`, e.g. ``'uint8'``.
        **This argument is mandatory in case** ``config`` **is** :obj:`False`.
        It is otherwise ignored.
      patches: The coordinates of the several patches to track, as an iterable
        (like a :obj:`list` or a :obj:`tuple`) containing one or several
        :obj:`tuple` of exactly :obj:`int` values. These integers correspond to
        the `y` position of the top-left corner of the patch, the `x` position
        of the top-left corner of the patch, the height of the patch, and the
        width of the patch. Up to 4 patches can be given and tracked. This
        argument must be provided if ``config`` is :obj:`False`.
      labels: The labels to use for sending data to downstream Blocks. If not
        given, the default labels are
        ``'t(s)', 'meta', 'Coord(px)', 'Eyy(%)', 'Exx(%)', 'Disp(px)'``. They
        carry for each image its timestamp, a :obj:`dict` containing its
        metadata, a :obj:`list` containing for each patch the coordinates of
        its center in a :obj:`tuple` of :obj:`int`, the `y` and `x` strain
        values calculated from the displacement and the initial position of the
        patches, and finally a :obj:`list` containing for each patch its
        displacement in the `y` and `x` direction in a :obj:`tuple` of
        :obj:`int`. If different labels are desired, they should all be
        provided at once in an iterable of :obj:`str` containing the correct
        number of labels (6).
      method: The method to use for performing the digital image correlation.
        Should be one of :
        ::

          `Disflow`, 'Pixel precision', 'Parabola', 'Lucas Kanade'

        ``'Disflow'`` uses OpenCV's DISOpticalFlow and ``'Lucas Kanade'`` uses
        OpenCV's calcOpticalFlowPyrLK, while all other methods are based on a
        basic cross-correlation in the Fourier domain. ``'Pixel precision'``
        calculates the displacement by getting the position of the maximum of
        the cross-correlation, and has thus a 1-pixel resolution. It is mainly
        meant for debugging. ``'Parabola'`` refines the result of
        ``'Pixel precision'`` by interpolating the neighborhood of the maximum,
        and has thus a sub-pixel resolution.
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`.
        Ignored if ``method`` is not ``'Disflow'``.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
        Ignored if ``method`` is not ``'Disflow'``.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`. Ignored if ``method`` is not ``'Disflow'``.
      finest_scale: Finest level of the Gaussian pyramid on which the flow is 
        computed in DISFlow, as an :obj:`int`. Zero level corresponds to the 
        original image resolution. The final flow is obtained by bilinear 
        upscaling. Ignored if ``method`` is not ``'Disflow'``.
      iterations: The number of fixed point iterations of variational 
        refinement per scale in DISFlow, as an :obj:`int`. Set to zero to 
        disable variational refinement completely. Higher values will typically 
        result in more smooth and high-quality flow. Ignored if ``method`` is
        not ``'Disflow'``.
      gradient_iterations: The maximum number of gradient descent iterations in
        the patch inverse search stage in DISFlow, as an :obj:`int`. Higher
        values may improve the quality. Ignored if ``method`` is not
        ``'Disflow'``.
      patch_size: The size of an image patch for matching in DISFlow, in pixels
        as an :obj:`int`. Ignored if ``method`` is not ``'Disflow'``.
      patch_stride: The stride between two neighbor patches in DISFlow, in
        pixels as an :obj:`int`. Must be less than the ``patch_size``. Lower
        values correspond to higher flow quality. Ignored if ``method`` is not
        ``'Disflow'``.
      border: The ratio of the patch that is kept for calculating the
        displacement, if ``method`` is ``'Disflow'``. For example if a value
        of `0.2` is given, only the center `80%` of the image is used for
        calculating the average displacement, in both directions. Ignored if
        ``method`` is not ``'Disflow'``.
      safe: If :obj:`True`, checks at each new image if the patches are not
        exiting the frame. Otherwise, the patches might exit the image which
        can lead to an unexpected behavior without raising an error.
      follow: If :obj:`True`, the position of each patch on the images is
        adjusted at each new image based on the previous computed displacement
        of this patch. If a displacement of 1 in the `x` direction was
        calculated on the previous image, and the patch is located at position
        `(x0, y0)`, the patch will be moved to position `(x0 + 1, y0)` for the
        next image. It "follows" the texture to track. Recommended if the
        expected displacement in pixels is big compared to the patch size. The
        only downside is that the patches may exit the frame if something goes
        wrong with the tracking.
      raise_on_patch_exit: If :obj:`True`, raises an exception when a tracked
        patch exits the border of the image, which stops the entire test.
        Otherwise, just logs a warning message and sleeps until the test is
        stopped in another way.
      **kwargs: Any additional argument will be passed to the
        :class:`~crappy.camera.Camera` object, and used as a kwarg to its
        :meth:`~crappy.camera.Camera.open` method.

    .. versionadded:: 1.5.7 *safe* and *follow* arguments
    .. versionremoved:: 1.5.9 *fields* argument
    .. versionadded:: 1.5.9 *method* argument
    .. versionadded:: 1.5.10 
       *transform*, *config*, *displayer_backend*, *displayer_framerate*, 
       *verbose*, *freq*, *save_images*, *img_name*, *save_folder*, 
       *save_period*, *save_backend* and *image_generator* arguments
    .. versionchanged:: 1.5.10 
       renamed *gditerations* argument to *gradient_iterations*
    .. versionchanged:: 1.5.10 
       renamed *show_image* argument to *display_images*
    .. versionadded:: 2.0.0 
       *debug*, *software_trig_label*, *img_extension*, *raise_on_patch_exit*,
       *img_size* and *img_dtype* arguments
    .. versionremoved:: 2.0.0 *img_name* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    if not config and patches is None:
      raise ValueError("If the config window is disabled, patches must be "
                       "provided !")

    super().__init__(camera=camera,
                     transform=transform,
                     config=config,
                     display_images=display_images,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=software_trig_label,
                     display_freq=display_freq,
                     freq=freq,
                     debug=debug,
                     save_images=save_images,
                     img_extension=img_extension,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     img_shape=img_shape,
                     img_dtype=img_dtype,
                     **kwargs)

    # Forcing the labels into a list
    if labels is None:
      self.labels = ['t(s)', 'meta', 'Coord(px)', 'Eyy(%)',
                     'Exx(%)', 'Disp(px)']
    elif isinstance(labels, str):
      self.labels = [labels]
    else:
      self.labels = list(labels)

    # Making sure a coherent number of labels and fields was given
    if len(self.labels) != 6:
      raise ValueError("The number of labels should be 6 !\n"
                       "Make sure that the time label was given")

    self._patches: Optional[SpotsBoxes] = None

    self._raise_on_exit = raise_on_patch_exit
    self._patches_int = list(patches) if patches is not None else None

    # These arguments are for the DICVEProcess
    self._method = method
    self._alpha = alpha
    self._delta = delta
    self._gamma = gamma
    self._finest_scale = finest_scale
    self._iterations = iterations
    self._gradient_iterations = gradient_iterations
    self._patch_size = patch_size
    self._patch_stride = patch_stride
    self._border = border
    self._safe = safe
    self._follow = follow

  def prepare(self) -> None:
    """This method mostly calls the :meth:`~crappy.blocks.Camera.prepare` 
    method of the parent class.
    
    In addition to that it instantiates the
    :class:`~crappy.blocks.camera_processes.DICVEProcess` object that performs
    the image correlation and the tracking.
    
    .. versionchanged:: 1.5.5 now accepting args and kwargs
    .. versionchanged:: 1.5.10 not accepting arguments anymore
    """

    # Instantiating the SpotsBoxes containing the patches to track
    self._patches = SpotsBoxes()
    if self._patches_int is not None:
      self._patches.set_spots(self._patches_int)
      self._patches.save_length()

    # Instantiating the DICVEProcess
    self.process_proc = DICVEProcess(
        patches=self._patches,
        method=self._method,
        alpha=self._alpha,
        delta=self._delta,
        gamma=self._gamma,
        finest_scale=self._finest_scale,
        iterations=self._iterations,
        gradient_iterations=self._gradient_iterations,
        patch_size=self._patch_size,
        patch_stride=self._patch_stride,
        border=self._border,
        safe=self._safe,
        follow=self._follow,
        raise_on_exit=self._raise_on_exit)

    super().prepare()

  def _configure(self) -> None:
    """This method should instantiate and start the
    :class:`~crappy.tool.camera_config.DICVEConfig` window for configuring the
    :class:`~crappy.camera.Camera` object.

    It should also handle the case when an exception is raised in the
    configuration window.
    """

    config = None

    # Instantiating and starting the configuration window
    try:
      config = DICVEConfig(self._camera, self._log_queue, self._log_level,
                           self.freq, self._patches)
      config.main()

    # If an exception is raised in the config window, closing it before raising
    except (Exception,) as exc:
      self._logger.exception("Caught exception in the configuration window !",
                             exc_info=exc)
      if config is not None:
        config.stop()
      raise CameraConfigError

    # Getting the image dtype and shape for setting the shared Array
    if config.shape is not None:
      self._img_shape = config.shape
    if config.dtype is not None:
      self._img_dtype = config.dtype



# coding: utf-8

from ...._global import DefinitionError


class MetaPath(type):
  """Metaclass ensuring that two Paths don't have the same name, and that all
  Paths define the required methods. Also keeps track of all the Path
  classes, including the custom user-defined ones.
  
  .. versionadded:: 2.0.0
  """

  classes = dict()

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that a Path with the same name doesn't already exist
    if name in cls.classes:
      raise DefinitionError(f"The {name} Generator Path is already defined !")

    # Saving the name
    cls.classes[name] = cls



# coding: utf-8

from time import time
from typing import List
import logging
from  warnings import warn

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class HandySens(InOut):
  """This class allows reading data from an OpSens HandySens fiber optics
  signal conditioner.

  It can read data from various fiber optics sensors like temperature,
  pressure, position or strain.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Opsens to HandySens
  """

  def __init__(self,
               device: str = '/dev/ttyUSB0') -> None:
    """Sets the argument and initializes the parent class.

    Args:
      device: Address of the serial connection for communicating with the
        OpSens.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._dev = None

    super().__init__()

    self._addr = device

  def open(self) -> None:
    """Opens the serial connection and configures the OpSens."""

    self.log(logging.INFO, f"Opening the serial connection on port "
                           f"{self._addr} with baudrate 57600")
    self._dev = serial.Serial(port=self._addr, baudrate=57600, timeout=0.1)
    self._send_cmd("meas:rate min")

  def get_data(self) -> List[float]:
    """Reads data from the OpSens and returns it."""

    return [time(), float(self._send_cmd("ch1:data? 1")[:-3])]

  def close(self) -> None:
    """Closes the serial connection, if it was opened."""

    if self._dev is not None:
      self.log(logging.INFO, f"Closing the serial connection on port "
                             f"{self._addr}")
      self._dev.close()

  def _send_cmd(self, cmd: str) -> str:
    """Wrapper for sending a command and returning the received answer."""

    self.log(logging.DEBUG, f"Writing b'{cmd}\\n' to port {self._addr}")
    self._dev.write(cmd + '\n')
    ret = self._dev.read_until(b'\x04\n').decode()
    self.log(logging.DEBUG, f"Read {ret} on port {self._addr}")
    return ret



# coding: utf-8

from typing import List, Optional, Iterable, Union
import tkinter as tk
import logging

from .meta_block import Block


class DashboardWindow(tk.Tk):
  """The GUI for displaying the label values.
  
  .. versionadded:: 1.5.7
  .. versionchanged:: 2.0.0 renamed from Dashboard_window to DashboardWindow
  """

  def __init__(self, labels: List[str]) -> None:
    """Initializes the GUI and sets the layout."""

    super().__init__()
    self.title('Dashboard')
    self.resizable(False, False)

    self._labels = labels

    # Attributes storing the tkinter objects
    self._tk_labels = {}
    self._tk_values = {}
    self.tk_var = {}

    # Setting the GUI
    self._set_variables()
    self._set_layout()

  def _set_variables(self) -> None:
    """Attributes one StringVar per label."""

    for label in self._labels:
      self.tk_var[label] = tk.StringVar(value='')

  def _set_layout(self) -> None:
    """Creates the Labels and places them on the GUI."""

    for row, label in enumerate(self._labels):
      # The name of the labels on the left
      self._tk_labels[label] = tk.Label(self, text=f'{label}:', borderwidth=15,
                                        font=("Courier bold", 48))
      self._tk_labels[label].grid(row=row, column=0)
      # Their values on the right
      self._tk_values[label] = tk.Label(self, borderwidth=15,
                                        textvariable=self.tk_var[label],
                                        font=("Courier bold", 48))
      self._tk_values[label].grid(row=row, column=1)


class Dashboard(Block):
  """This Block generates an interface displaying data as text in a dedicated
  window.

  It relies on a :obj:`~tkinter.Tk` window for the graphical interface.

  In the window, the left column contains the names of the labels to display
  and the right column contains the latest received values for these labels.
  For each label, only the last value is therefore displayed.

  This Block provides a nicer display than the raw
  :class:`~crappy.blocks.LinkReader` Block. For displaying the evolution of a
  label over time, the :class:`~crappy.blocks.Grapher` Block should be used
  instead.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               labels: Union[str, Iterable[str]],
               nb_digits: int = 2,
               display_freq: bool = False,
               freq: Optional[float] = 30,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      labels: Only the data from these labels will be displayed on the window.
      nb_digits: Number of decimals to show.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 1.5.7 *verbose* and *freq* arguments
    .. versionadded:: 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._dashboard: Optional[DashboardWindow] = None

    super().__init__()
    self.display_freq = display_freq
    self.freq = freq
    self.debug = debug

    self._labels = [labels] if isinstance(labels, str) else list(labels)
    self._nb_digits = nb_digits

  def prepare(self) -> None:
    """Checks that there's at least one incoming :class:`~crappy.links.Link`,
    and starts the GUI.
    
    .. versionadded:: 1.5.7
    """

    if not self.inputs:
      raise IOError("No Link pointing towards the Dashboard Block !")

    self.log(logging.INFO, "Creating the dashboard window")
    self._dashboard = DashboardWindow(self._labels)
    self._dashboard.update()

  def loop(self) -> None:
    """Receives the data from the incoming :class:`~crappy.links.Link` and
    displays it.
    
    .. versionadded:: 1.5.7
    """

    data = self.recv_last_data(fill_missing=False)

    for label, value in data.items():
      # Only displays the required labels
      if label in self._labels:
        # Possibility to display str values carried by the links
        if isinstance(value, str):
          self.log(logging.DEBUG, f"Displaying {value} for the label {label} "
                                  f"on the dashboard")
          self._dashboard.tk_var[label].set(value)
        elif isinstance(value, int) or isinstance(value, float):
          self.log(logging.DEBUG, f"Displaying {value:.{self._nb_digits}f} for"
                                  f" the label {label} on the dashboard")
          self._dashboard.tk_var[label].set(f'{value:.{self._nb_digits}f}')

    # In case the GUI has been destroyed, don't raise an error
    try:
      self._dashboard.update()
    except tk.TclError:
      pass

  def finish(self) -> None:
    """Closes the display.

    .. versionadded:: 1.5.7
    """

    # In case the GUI has been destroyed, don't raise an error
    try:
      if self._dashboard is not None:
        self.log(logging.INFO, "Closing the dashboard window")
        self._dashboard.destroy()
    except tk.TclError:
      pass



# coding: utf-8

import tkinter as tk
from tkinter.messagebox import showerror
from typing import Optional
import numpy as np
from io import BytesIO
from pkg_resources import resource_string
from time import sleep
import logging
from multiprocessing.queues import Queue

from .camera_config_boxes import CameraConfigBoxes
from .config_tools import Box, SpotsDetector
from ...camera.meta_camera import Camera
from ..._global import OptionalModule

try:
  from PIL import Image
except (ModuleNotFoundError, ImportError):
  Image = OptionalModule("pillow")


class VideoExtensoConfig(CameraConfigBoxes):
  """Class similar to :class:`~crappy.tool.camera_config.CameraConfig` but also
  displaying the bounding boxes of the detected spots, and allowing to select
  the area where to detect the spots by drawing a box with the left mouse
  button.

  It relies on the :class:`~crappy.tool.camera_config.config_tools.Box` and
  :class:`~crappy.tool.camera_config.config_tools.SpotsDetector` tools. It is
  meant to be used for configuring the :class:`~crappy.blocks.VideoExtenso`
  Block.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from VE_config to VideoExtensoConfig
  """

  def __init__(self,
               camera: Camera,
               log_queue: Queue,
               log_level: Optional[int],
               max_freq: Optional[float],
               detector: SpotsDetector) -> None:
    """Sets the args and initializes the parent class.

    Args:
      camera: The :class:`~crappy.camera.Camera` object in charge of acquiring
      the images.
      log_queue: A :obj:`multiprocessing.Queue` for sending the log messages to
        the main :obj:`~logging.Logger`, only used in Windows.
      log_level: The minimum logging level of the entire Crappy script, as an
        :obj:`int`.
      max_freq: The maximum frequency this window is allowed to loop at. It is
        simply the ``freq`` attribute of the :class:`~crappy.blocks.Camera`
        Block.
      detector: An instance of
        :class:`~crappy.tool.camera_config.config_tools.SpotsDetector` used for
        detecting spots on the images received from the
        :class:`~crappy.camera.Camera`.

    .. versionchanged:: 1.5.10 renamed *ve* argument to *video_extenso*
    .. versionadded:: 2.0.0
       *detector*, *log_queue*, *log_level* and *max_freq* arguments
    .. versionremoved:: 2.0.0 *video_extenso* argument
    """

    super().__init__(camera, log_queue, log_level, max_freq)
    self._detector = detector
    self._spots = detector.spots

  def finish(self) -> None:
    """Method called when the user tries to close the configuration window.

    Checks that spots were detected on the image. If not, warns the user and
    prevents him from exiting except with CTRL+C. Also, saves the initial
    length if not already done by the user.
    
    .. versionadded:: 2.0.0
    """

    if self._detector.spots.empty():
      self.log(logging.WARNING, "No spots were selected ! Not exiting the "
                                "configuration window")
      showerror("Error !",
                message="Please select spots before exiting the config "
                        "window !\nOr hit CTRL+C to exit Crappy")
      return

    if self._detector.spots.x_l0 is None or self._detector.spots.y_l0 is None:
      self._detector.spots.save_length()
      self.log(logging.INFO,
               f"Successfully saved L0 ! L0 x : {self._detector.spots.x_l0}, "
               f"L0 y : {self._detector.spots.y_l0}")

    super().stop()

  def _set_bindings(self) -> None:
    """Binds the left mouse button click for drawing the box in which the spots
    will be searched."""

    super()._set_bindings()

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _create_buttons(self) -> None:
    """Compared with the parent class, creates an extra button for saving the
    original position of the spots."""

    self._update_button = tk.Button(self._sets_frame, text="Apply Settings",
                                    command=self._update_settings)
    self._update_button.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

    self._update_button = tk.Button(self._sets_frame, text="Save L0",
                                    command=self._save_l0)
    self._update_button.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

  def _stop_box(self, _: tk.Event) -> None:
    """When the user releases the mouse, searches for spots in the selected
    area and displays them if any were found."""

    # If it's just a regular click with no dragging, do nothing
    if self._img is None or self._select_box.no_points():
      self._select_box.reset()
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = self._select_box.sorted()

    # If the box is flat, resetting it
    if y_left == y_right or x_top == x_bottom:
      self._select_box.reset()
      return

    # Now actually trying to detect the spots
    try:
      self._detector.detect_spots(self._original_img[x_top: x_bottom,
                                                     y_left: y_right],
                                  x_top, y_left)
    except IndexError:
      # Highly unlikely but always better to be careful
      self._detector.spots.reset()
      return

    # This box is not needed anymore
    self._select_box.reset()

  def _save_l0(self) -> None:
    """Saves the original positions of the spots on the image."""

    if self._detector.spots.empty():
      self.log(logging.WARNING, "Cannot save L0, there are no spots !")
    else:
      self._detector.spots.save_length()
      self.log(logging.INFO,
               f"Successfully saved L0 ! L0 x : {self._detector.spots.x_l0}, "
               f"L0 y : {self._detector.spots.y_l0}")

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the patches and the
    select box on top of the displayed image."""

    self.log(logging.DEBUG, "The image canvas was resized")

    self._draw_box(self._select_box)
    self._draw_spots()
    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the patches and the
    select box on top of the displayed image."""

    self.log(logging.DEBUG, "Updating the image")

    ret = self._camera.get_image()

    # If no frame could be grabbed from the camera
    if ret is None:
      # If it's the first call, generate error image to initialize the window
      if not self._n_loops:
        self.log(logging.WARNING, "Could not get an image from the camera, "
                                  "displaying an error image instead")
        ret = None, np.array(Image.open(BytesIO(resource_string(
          'crappy', 'tool/data/no_image.png'))))
      # Otherwise, just pass
      else:
        self.log(logging.DEBUG, "No image returned by the camera")
        self.update()
        sleep(0.001)
        return

    self._n_loops += 1
    _, img = ret

    if img.dtype != self.dtype:
      self.dtype = img.dtype
    if self.shape != img.shape:
      self.shape = img.shape

    self._cast_img(img)
    self._draw_box(self._select_box)
    self._draw_spots()
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, _: Box) -> None:
    """If a patch is outside the image, it means that the image size has been
    modified. Simply resetting the spots then."""

    self._spots.reset()



# coding: utf-8

import numpy as np
from typing import Union, Optional
from pathlib import Path
import logging

from .._global import OptionalModule
from .meta_block import Block

try:
  import tables
except ModuleNotFoundError:
  tables = OptionalModule("tables", "HDFRecorder needs the tables module to "
                          "write hdf files.")


class HDFRecorder(Block):
  """This Block records data efficiently into a HDF5 file.
  
  It expects data as :obj:`numpy.array` from exactly one upstream Block, that
  should be an :class:`~crappy.blocks.IOBlock` in `streamer` mode. It then 
  saves this data in a HDF5 file using the :mod:`tables` module.
  
  This Block is intended for high-speed data recording from 
  :class:`~crappy.inout.InOut` in `streamer` mode. For regular data recording,
  the :class:`~crappy.blocks.Recorder` Block should be used instead.
  
  Warning:
    Corrupted HDF5 files are not readable at all ! If anything goes wrong 
    during a test, especially during the finish phase, it is not guaranteed 
    that the recorded data will be readable.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Hdf_recorder to HDFRecorder
  """

  def __init__(self,
               filename: Union[str, Path],
               node: str = 'table',
               expected_rows: int = 10**8,
               atom=None,
               label: str = 'stream',
               metadata: Optional[dict] = None,
               freq: Optional[float] = None,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      filename: Path to the output file, either relative or absolute. If the
        parent folders of the file do not exist, they will be created. If the
        file already exists, the actual file where data will be written will be
        renamed with a trailing index to avoid overriding it.
      node: The name of the array to create in the HDF5 file, as a :obj:`str`.
      expected_rows: The number of expected rows in the file. It is used to
        optimize the dumping.
      atom: This represents the type of data to be stored in the table. It can
        be given as a :obj:`tables.Atom` instance, as a :obj:`numpy.array`
        or as a :obj:`str`.
      label: The label carrying the data to be saved
      metadata: A :obj:`dict` containing additional information to save in the
        `HDF5` file.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: if :obj:`True`, displays the looping frequency of the 
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded:: 1.5.10 *freq* and *verbose* arguments
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._hfile = None

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._path = Path(filename)
    self._label = label
    self._metadata = {} if metadata is None else metadata
    self._expected_rows = expected_rows

    self._node = node
    atom = tables.Int16Atom() if atom is None else atom
    if not isinstance(atom, tables.Atom):
      self._atom = tables.Atom.from_dtype(np.dtype(atom))
    else:
      self._atom = atom

    self._array_initialized = False

  def prepare(self) -> None:
    """Checks that the Block has the right number of inputs, creates the
    folder containing the file if it doesn't already exist, changes the name of
    the file if it already exists, and initializes the HDF file."""

    # Making sure there's the right number of incoming links
    if not self.inputs:
      raise ValueError('The HDF Recorder block does not have inputs !')
    elif len(self.inputs) > 1:
      raise ValueError('Cannot link more than one block to an HDF Recorder '
                       'block !')

    parent_folder = self._path.parent

    # Creating the folder for storing the data if it does not already exist
    if not Path.is_dir(parent_folder):
      self.log(logging.INFO, f"Creating the folder containing the file to save"
                             f" data to ({parent_folder})")
      Path.mkdir(parent_folder, exist_ok=True, parents=True)

    # Changing the name of the file if it already exists
    if Path.exists(self._path):
      self.log(logging.WARNING, f"The file {self._path} already exists !")
      stem, suffix = self._path.stem, self._path.suffix
      i = 1
      # Adding an integer at the end of the name to identify the file
      while Path.exists(parent_folder / f'{stem}_{i:05d}{suffix}'):
        i += 1
      self._path = parent_folder / f'{stem}_{i:05d}{suffix}'
      self.log(logging.WARNING, f"Writing data to the file {self._path} "
                                f"instead !")

    # Initializing the file to save data to
    self.log(logging.INFO, "Initializing the HDF5 file")
    self._hfile = tables.open_file(str(self._path), "w")
    for name, value in self._metadata.items():
      self._hfile.create_array(self._hfile.root, name, value)

  def loop(self) -> None:
    """Receives data from the upstream Block and saves it.

    Also creates the array for recording data when the first values are
    received.
    """

    # Do nothing until the first value to save are received
    if not self._array_initialized:
      if self.data_available():
        self._first_loop()
        self._array_initialized = True
      else:
        return

    data = self.recv_all_data()

    if self._label in data:
      for elt in data[self._label]:
        self._array.append(elt)

  def finish(self) -> None:
    """Closes the HDF file."""

    if self._hfile is not None:
      self.log(logging.INFO, "Closing the HDF5 file")
      self._hfile.close()

  def _first_loop(self) -> None:
    """Initializes the array for saving data."""

    data = self.recv_all_data()

    if self._label not in data:
      raise KeyError(f'The data received by the HDF Recorder block does not '
                     f'contain the label {self._label} !')

    self.log(logging.INFO, "Initializing the arrays in the HDF5 file")

    _, width, *_ = data[self._label][0].shape
    self._array = self._hfile.create_earray(self._hfile.root,
                                            self._node,
                                            self._atom,
                                            (0, width),
                                            expectedrows=self._expected_rows)
    for elt in data[self._label]:
      self._array.append(elt)



# coding: utf-8

from typing import Union, Dict
import logging

from .meta_path import Path, ConditionType


class Constant(Path):
  """The simplest Path, outputs the same constant value until the stop 
  condition is met.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               condition: Union[str, ConditionType],
               value: float = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      condition: The condition for switching to the next Path. Refer to
        :class:`~crappy.blocks.generator_path.meta_path.Path` for more
        information.
      value: The value to output.
      
    .. versionchanged:: 1.5.10 renamed *time* argument to *_last_time*
    .. versionchanged:: 1.5.10 renamed *cmd* argument to *_last_cmd*
    .. versionremoved:: 1.5.10 *send_one* argument
    .. versionremoved:: 2.0.0 *_last_time* and *_last_cmd* arguments
    """

    super().__init__()

    self._condition = self.parse_condition(condition)

    if value is None and self.last_cmd is None:
      raise ValueError('For the first path, a value must be given !')

    self._value = self.last_cmd if value is None else value

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met.
    
    .. versionadded:: 1.5.10
    """

    # Checking if the stop condition is met
    if self._condition(data):
      self.log(logging.DEBUG, "Stop condition met")
      raise StopIteration

    # Returning the value
    return self._value



# coding: utf-8

from ..._global import DefinitionError


class MetaBlock(type):
  """Metaclass ensuring that two Blocks don't have the same name, and that all
  Blocks define the required methods. Also keeps track of all the Block
  classes, including the custom user-defined ones.

  .. versionadded:: 2.0.0
  """

  existing = list()

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that a Block with the same name doesn't already exist
    if name in cls.existing:
      raise DefinitionError(f"The {name} Block is already defined !")

    # Saving the name
    cls.existing.append(name)



# coding: utf-8

from __future__ import annotations
from typing import Tuple, Optional, Callable, Union
from re import findall, search, finditer, split, Match
from dataclasses import dataclass
import logging
from subprocess import run
from multiprocessing import current_process


@dataclass
class V4L2Parameter:
  """A class for the different parameters the user can adjust."""

  name: str
  type: str
  min: Optional[str] = None
  max: Optional[str] = None
  step: Optional[str] = None
  default: Optional[str] = None
  value: Optional[str] = None
  flags: Optional[str] = None
  options: Optional[Tuple[str, ...]] = None

  # Regex to extract the different parameters and their information
  param_pattern = (r'(\w+)\s+0x\w+\s+\((\w+)\)\s+:\s*'
                   r'(min=(-?\d+)\s+)?'
                   r'(max=(-?\d+)\s+)?'
                   r'(step=(\d+)\s+)?'
                   r'(default=(-?\d+)\s+)?'
                   r'value=(-?\d+)\s*'
                   r'(flags=([^\\n]+))?')

  option_pattern = r'(\w+ \w+ \(menu\))([\s\S]+?)(?=\n\s*\w+ \w+ \(.+?\)|$)'

  @classmethod
  def parse_info(cls, match: Match) -> V4L2Parameter:
    """Instantiates the class Parameter, according to the information
     collected with v4l2-ctl.

    Args:
      match: Match object returned by successful matches of the regex with
      a string.

    Returns:
      The instantiated class.
    """

    return cls(name=match.group(1),
               type=match.group(2),
               min=match.group(4) if match.group(4) else None,
               max=match.group(6) if match.group(6) else None,
               step=match.group(8) if match.group(8) else None,
               default=match.group(10) if match.group(10) else None,
               value=match.group(11),
               flags=match.group(13) if match.group(13) else None)

  def add_options(self, match: Match) -> None:
    """Adds the different possible options for a menu parameter.

    Args:
      match: Match object returned by successful matches of the regex with
      a string.
    """

    menu_info = match.group(1)
    menu_values = match.group(2)
    menu_name = search(r'(\w+) \w+ \(menu\)', menu_info).group(1)
    if self.name == menu_name:
      options = findall(r'\d+: .+?(?=\n|$)', menu_values)
      num_options = findall(r'(\d+): .+?(?=\n|$)', menu_values)
      self.options = tuple(options)
      for num, opt in zip(num_options, options):
        if self.default == num:
          self.default = opt


class V4L2Helper:
  """A class for getting parameters available in a camera by using
  v4l-utils.

  .. versionadded:: 2.0.0
  """

  def __init__(self):
    """Simply initializes the instance attributes."""

    self._parameters = list()
    self._formats = list()
    self._logger: Optional[logging.Logger] = None

  def _get_param(self, device: Optional[Union[str, int]]) -> None:
    """Extracts the different parameters and their information by parsing
    v4l2-ctl with regex."""

    # Trying to run v4l2-ctl to get the available settings
    if device is None:
      command = ['v4l2-ctl', '-L']
    else:
      command = ['v4l2-ctl', '-d', str(device), '-L']
    self.log(logging.DEBUG, f"Getting the available image settings with "
                            f"command {' '.join(command)}")
    ret = run(command, capture_output=True, text=True).stdout
    self.log(logging.DEBUG, f"Got the following image settings: {ret}")

    # Extract the different parameters and their information
    matches = finditer(V4L2Parameter.param_pattern, ret)
    for match in matches:
      self._parameters.append(V4L2Parameter.parse_info(match))

    # Regex to extract the different options in a menu
    menu_options = finditer(V4L2Parameter.option_pattern, ret)

    # Extract the different options
    for menu_option in menu_options:
      for param in self._parameters:
        param.add_options(menu_option)

  def _get_available_formats(self, device: Optional[Union[str, int]]) -> None:
    """Extracts the different formats available by parsing v4l2-ctl with
    regex."""

    # Trying to run v4l2-ctl to get the available formats
    if device is None:
      command = ['v4l2-ctl', '--list-formats-ext']
    else:
      command = ['v4l2-ctl', '-d', str(device), '--list-formats-ext']
    self.log(logging.DEBUG, f"Getting the available image formats with "
                            f"command {' '.join(command)}")
    ret = run(command, capture_output=True, text=True).stdout
    self.log(logging.DEBUG, f"Got the following image formats: {ret}")

    # Splitting the returned string to isolate each encoding
    if findall(r'\[\d+]', ret):
      formats = split(r'\[\d+]', ret)[1:]
    elif findall(r'Pixel\sFormat', ret):
      formats = split(r'Pixel\sFormat', ret)[1:]
    else:
      formats = list()

    # For each encoding, finding its name, available sizes and framerates
    if formats:
      for img_format in formats:
        name, *_ = search(r"'(\w+)'", img_format).groups()
        sizes = findall(r'\d+x\d+', img_format)
        fps_sections = split(r'\d+x\d+', img_format)[1:]

        # Formatting the detected sizes and framerates into strings
        for size, fps_section in zip(sizes, fps_sections):
          fps_list = findall(r'\((\d+\.\d+)\sfps\)', fps_section)
          for fps in fps_list:
            self._formats.append(f'{name} {size} ({fps} fps)')

  def _add_setter(self,
                  name: str,
                  device: Optional[Union[int, str]]) -> Callable:
    """Creates a setter function for a setting named 'name'.

    Args:
      name: Name of the setting.

    Returns:
      The setter function.
    """

    def setter(value: Union[str, int, bool]) -> None:
      """The method to set the value of a setting running v4l2-ctl."""

      if isinstance(value, str):
        # The value to set the menu parameter is just the int
        # at the beginning the string
        value = search(r'(\d+): ', value).group(1)
        if device is not None:
          command = ['v4l2-ctl', '-d', str(device), '--set-ctrl',
                     f'{name}={value}']
        else:
          command = ['v4l2-ctl', '--set-ctrl', f'{name}={int(value[0])}']
        self.log(logging.DEBUG, f"Running the command {' '.join(command)}")
        run(command, capture_output=True, text=True)
        self.log(logging.DEBUG, f"Set {name} to {int(value[0])}")
      else:
        if device is not None:
          command = ['v4l2-ctl', '-d', str(device), '--set-ctrl',
                     f'{name}={int(value)}']
        else:
          command = ['v4l2-ctl', '--set-ctrl', f'{name}={int(value)}']
        self.log(logging.DEBUG, f"Running the command {' '.join(command)}")
        run(command, capture_output=True, text=True)
        self.log(logging.DEBUG, f"Set {name} to {int(value)}")
    return setter

  def _add_scale_getter(self,
                        name: str,
                        device: Optional[Union[int, str]]) -> Callable:
    """Creates a getter function for a setting named 'name'.

    Args:
      name: Name of the setting.

    Returns:
      The getter function.
    """

    def getter() -> int:
      """The method to get the current value of a scale setting running
       v4l2-ctl."""

      # Trying to run v4l2-ctl to get the value
      if device is not None:
        command = ['v4l2-ctl', '-d', str(device), '--get-ctrl', name]
      else:
        command = ['v4l2-ctl', '--get-ctrl', name]
      self.log(logging.DEBUG, f"Running the command {' '.join(command)}")
      value = run(command, capture_output=True, text=True).stdout
      value = int(search(r':\s(-?\d+)', value).group(1))
      self.log(logging.DEBUG, f"Got {name}: {value}")
      return value
    return getter

  def _add_bool_getter(self,
                       name: str,
                       device: Optional[Union[int, str]]) -> Callable:
    """Creates a getter function for a setting named 'name'.

    Args:
      name: Name of the setting.

    Returns:
      The getter function.
    """

    def getter() -> bool:
      """The method to get the current value of a bool setting running
      v4l2-ctl."""

      # Trying to run v4l2-ctl to get the value
      if device is not None:
        command = ['v4l2-ctl', '-d', str(device), '--get-ctrl', name]
      else:
        command = ['v4l2-ctl', '--get-ctrl', name]
      self.log(logging.DEBUG, f"Running the command {' '.join(command)}")
      value = run(command, capture_output=True, text=True).stdout
      value = bool(int(search(r':\s(\d+)', value).group(1)))
      self.log(logging.DEBUG, f"Got {name}: {value}")
      return value
    return getter

  def _add_menu_getter(self,
                       name: str,
                       device: Optional[Union[int, str]]) -> Callable:
    """Creates a getter function for a setting named 'name'.

    Args:
      name: Name of the setting.

    Returns:
      The getter function.
    """

    def getter() -> str:
      """The method to get the current value of a choice setting
      running v4l2-ctl."""

      # Trying to run v4l2-ctl to get the value
      if device is not None:
        command = ['v4l2-ctl', '-d', str(device), '--get-ctrl', name]
      else:
        command = ['v4l2-ctl', '--get-ctrl', name]
      self.log(logging.DEBUG, f"Running the command {' '.join(command)}")
      value = run(command, capture_output=True, text=True).stdout
      value = search(r':\s(\d+)', value).group(1)
      for param in self._parameters:
        if param.name == name:
          for option in param.options:
            if value == search(r'(\d+):', option).group(1):
              value = option
      self.log(logging.DEBUG, f"Got {name}: {value}")
      return value
    return getter

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the Modifiers.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)



# coding: utf-8

from time import time
from typing import Tuple, Optional
import numpy as np
import logging

from .meta_camera import Camera


class FakeCamera(Camera):
  """This camera class generates images without requiring any actual camera or
  existing image file.

  The generated images are just a gradient of grey levels, with a line moving
  as a function of time. It is possible to tune the dimension of the image, the
  frame rate and the speed of the moving line.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Fake_camera to FakeCamera
  """

  def __init__(self) -> None:
    """Initializes the parent class and instantiates the settings."""

    super().__init__()
    self._frame_nr = -1

    self.add_scale_setting('width', 1, 4096, None, self._gen_image, 1280, 1)
    self.add_scale_setting('height', 1, 4096, None, self._gen_image, 720, 1)
    self.add_scale_setting('speed', 0., 800., None, None, 400., 0.8)
    self.add_scale_setting('fps', 0.1, 100., None, None, 50., 0.1)

    self._t0 = time()
    self._t = self._t0

  def open(self,
           width: int = 1280,
           height: int = 720,
           speed: float = 100.,
           fps: float = 50.) -> None:
    """Sets the settings, initializes the first image and starts the time
    counter.

    Args:
      width: The width of the image to generate in pixels.
      height: The height of the image to generate in pixels.
      speed: The evolution speed of the image, in pixels per second.
      fps: The maximum update frequency of the generated images.
    
    .. versionadded:: 2.0.0
       *width*, *height*, *speed* and *fps* arguments explicitly listed
    """

    self.set_all(width=width, height=height, speed=speed, fps=fps)

    self._gen_image()

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Returns the updated image, depending only on the current timestamp.

    Also includes a waiting loop in order to achieve the right frame rate.
    """

    # Waiting in order to achieve the right frame rate
    while time() - self._t < 1 / self.fps:
      pass

    self._t = time()
    self._frame_nr += 1

    # Splitting the image to make a moving line
    row = int(self.speed * (self._t - self._t0)) % self.height
    return self._t, np.concatenate((self._img[row:], self._img[:row]), axis=0)

  def _gen_image(self, _: Optional[float] = None) -> None:
    """Generates the base gradient image, that will be split and returned
    in the :meth:`get_image` method."""

    self.log(logging.DEBUG, "Generating the image")
    self._img = np.arange(self.height) * 255. / self.height
    self._img = np.repeat(self._img.reshape(self.height, 1),
                          self.width, axis=1).astype(np.uint8)



# coding: utf-8

from typing import Iterable, Optional, Union
from pathlib import Path
import logging

from .meta_block import Block


class Recorder(Block):
  """This Block saves data from an upstream Block to a text file, with values 
  separated by a coma and lines by a newline character.

  The first row of the file contains the names of the saved labels. This Block 
  can only save data coming from exactly one upstream Block. To save data
  from multiple Blocks, use several instances of Recorder (recommended) or a
  :class:`~crappy.blocks.Multiplexer` Block.
  
  This Block cannot directly record data from "streams", i.e. coming from an
  :class:`~crappy.blocks.IOBlock` Block with the ``'streamer'`` argument set to
  :obj:`True`. To do so, the :class:`~crappy.blocks.HDFRecorder` Block should
  be used instead. Alternatively, a :class:`~crappy.modifier.Demux` Modifier 
  can be placed between the IOBlock and the Recorder, but most of the acquired
  data won't be saved.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               file_name: Union[str, Path],
               delay: float = 2,
               labels: Optional[Union[str, Iterable[str]]] = None,
               freq: Optional[float] = 200,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      file_name: Path to the output file, either relative or absolute. If the
        parent folders of the file do not exist, they will be created. If the
        file already exists, the actual file where data will be written will be
        renamed with a trailing index to avoid overriding it.
      delay: Delay between each write in seconds.
      labels: If provided, only the data carried by these labels will be saved.
        Otherwise, all the received data is saved.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the 
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded 1.5.10 *freq* and *verbose* arguments
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    .. versionchanged:: 2.0.0 renamed *filename* argument to *file_name*
    """

    super().__init__()
    self.niceness = -5
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    self._delay = delay
    self._path = Path(file_name)

    # Forcing the labels into a list
    if labels is not None and isinstance(labels, str):
      self._labels = [labels]
    elif labels is not None:
      self._labels = list(labels)
    else:
      self._labels = None

    self._file_initialized = False

  def prepare(self) -> None:
    """Checks that the Block has the right number of inputs, creates the
    folder containing the file if it doesn't already exist, and changes the
    name of the file if it already exists."""

    # Making sure there's the right number of incoming links
    if not self.inputs:
      raise ValueError('The Recorder block does not have inputs !')
    elif len(self.inputs) > 1:
      raise ValueError('Cannot link more than one block to a Recorder block !')

    parent_folder = self._path.parent

    # Creating the folder for storing the data if it does not already exist
    if not Path.is_dir(parent_folder):
      self.log(logging.INFO, f"Creating the folder containing the file to save"
                             f" data to ({parent_folder})")
      Path.mkdir(parent_folder, exist_ok=True, parents=True)

    # Changing the name of the file if it already exists
    if Path.exists(self._path):
      self.log(logging.WARNING, f"The file {self._path} already exists !")
      stem, suffix = self._path.stem, self._path.suffix
      i = 1
      # Adding an integer at the end of the name to identify the file
      while Path.exists(parent_folder / f'{stem}_{i:05d}{suffix}'):
        i += 1
      self._path = parent_folder / f'{stem}_{i:05d}{suffix}'
      self.log(logging.WARNING, f"Writing data to the file {self._path} "
                                f"instead !")

  def loop(self) -> None:
    """Receives data from the upstream Block and saves it."""

    if not self._file_initialized:
      if self.data_available():

        data = self.recv_all_data(delay=self._delay)

        # If no labels are given, save everything that's received
        if self._labels is None:
          self._labels = list(data.keys())

        # The first row of the file contains the names of the labels
        with open(self._path, 'w') as file:
          self.log(logging.INFO, f"Writing the header on file {self._path}")
          file.write(f"{','.join(self._labels)}\n")

        self._file_initialized = True
      else:
        return

    else:
      data = self.recv_all_data(delay=self._delay)

    # Keeping only the data that needs to be saved
    data = {key: val for key, val in data.items() if key in self._labels}

    if data:
      with open(self._path, 'a') as file:
        # Sorting the lists of values in the same order as the labels
        sorted_data = [data[label] for label in self._labels]
        # Actually writing the values
        self.log(logging.DEBUG, f"Writing {sorted_data} to the file "
                                f"{self._path}")
        for values in zip(*sorted_data):
          file.write(f"{','.join(map(str, values))}\n")
